def factorial(n):
    res = 1
    for i in range(2, n + 1):
        res *= i
    return res


def binomial(n, k):
    if k > n:
        return 0
    if n - k < k:
        k = n - k
    res = 1
    for i in range(1, k + 1):
        res = res * (n - (k - i)) // i
    return res


def catalan(n):
    return binomial(2 * n, n) // (n + 1)



import collections
import numbers
import random
from itertools import cycle, islice

from .interfaces.flippable import Flippable
from .interfaces.patt import Patt
from .interfaces.rotatable import Rotatable
from .interfaces.shiftable import Shiftable
from .misc import DIR_EAST, DIR_NONE, DIR_NORTH, DIR_SOUTH, DIR_WEST
from .perm import Perm

MeshPatternBase = collections.namedtuple("MeshPatternBase", ["pattern", "shading"])


class MeshPatt(MeshPatternBase, Patt, Rotatable, Shiftable, Flippable):
    """A mesh pattern class.

    Attributes:
        pattern: <permuta.Perm>
            The underlying classical pattern.
        shading: frozenset
            The shading as a immutable set of coordinates, with lower-left as
            origin.
    """

    def __new__(cls, pattern=Perm(), shading=frozenset()):
        """Return a MeshPatt instance.

        Args:
            cls:
                The class of which an instance is requested.
            pattern: <permuta.Perm> or <collections.Iterable>
                An perm or an iterable corresponding to a legal perm.
            shading: <collections.Iterable>
                An iterable of 2-tuples.
        Raises:
            TypeError:
                Bad argument type.
            ValueError:
                Bad argument, but correct type.
        """
        if not isinstance(pattern, Perm):
            pattern = Perm(pattern)
        if not isinstance(shading, frozenset):
            shading = frozenset(shading)
        return super(MeshPatt, cls).__new__(cls, pattern, shading)

    def __init__(self, pattern=Perm(), shading=frozenset()):
        for coordinate in self.shading:
            if not isinstance(coordinate, tuple):
                message = "'{}' object is not a tuple".format(repr(coordinate))
                raise TypeError(message)
            if len(coordinate) != 2:
                message = "Element is not a shading coordinate: '{}'".format(
                    repr(coordinate)
                )
                raise ValueError(message)
            x, y = coordinate
            if not isinstance(x, numbers.Integral):
                message = "'{}' object is not an integer".format(repr(x))
                raise TypeError(message)
            if not isinstance(y, numbers.Integral):
                message = "'{}' object is not an integer".format(repr(y))
                raise TypeError(message)
            if (not 0 <= x <= len(self.pattern)) or (not 0 <= y <= len(self.pattern)):
                message = "Element out of range: '{}'".format(coordinate)
                raise ValueError(message)

    #
    # Methods returning new permutations
    #

    def complement(self):
        """Returns the complement of the mesh pattern, which has the complement
        of the underlying pattern and every shading flipped across the
        horizontal axis.

        Returns: <permuta.MeshPatt>
            The complement of the meshpatt.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1)})).complement()
            MeshPatt(Perm((0,)), [(0, 0)])
            >>> MeshPatt(Perm((0, 2, 1)),
            ... frozenset({(0, 1), (0, 2), (0, 3)})).complement()
            MeshPatt(Perm((2, 0, 1)), [(0, 0), (0, 1), (0, 2)])
        """
        return MeshPatt(
            self.pattern.complement(), [(x, len(self) - y) for (x, y) in self.shading]
        )

    def reverse(self):
        """Returns the reversed mesh patterns, which has the underlying pattern
        reversed and every shading flipped across the vertical axis.

        Returns: <permuta.MeshPatt>
            The meshpatt reversed.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1)})).reverse()
            MeshPatt(Perm((0,)), [(1, 1)])
            >>> MeshPatt(Perm((2, 1, 0)),
            ... frozenset({(3, 2), (2, 2), (1, 1)})).reverse()
            MeshPatt(Perm((0, 1, 2)), [(0, 2), (1, 2), (2, 1)])
        """
        return MeshPatt(
            self.pattern.reverse(), [(len(self) - x, y) for (x, y) in self.shading]
        )

    def inverse(self):
        """Returns the inverse of the meshpatt, that is the meshpatt with the
        underlying classical pattern as the inverse and the shadings hold the
        same relation between the points. This is equivalent to flipping the
        pattern over the diagonal.

        Returns: <permuta.MeshPatt>
            The 'inverse' of the meshpatt.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1)})).inverse()
            MeshPatt(Perm((0,)), [(1, 0)])
        """
        return MeshPatt(self.pattern.inverse(), [(y, x) for (x, y) in self.shading])

    def sub_mesh_pattern(self, indices):
        """Return the mesh pattern induced by indices.

        Args:
            indices: <collections.Iterable> of <numbers.Integral>
                A list of unique indices of elements in self.

        Returns: <permuta.MeshPattern>
            A mesh pattern where the pattern is the permutation induced by the
            indices and a region is shaded if and only if the corresponding
            region of self is fully shaded.

        Exampes:
            >>> shading = frozenset({(3, 2), (1, 3), (4, 2), (0, 3), (1, 2),
            ... (4, 3)})
            >>> MeshPatt(Perm((3, 2, 1, 0)),
            ... shading).sub_mesh_pattern((0, 1, 3))
            MeshPatt(Perm((2, 1, 0)), [(0, 2), (1, 2), (3, 2)])
            >>> MeshPatt(Perm((2, 3, 1, 0)),
            ... shading).sub_mesh_pattern((1, 2, 3))
            MeshPatt(Perm((2, 1, 0)), [(3, 2)])
        """
        indices = sorted(indices)
        if not indices:
            return MeshPatt()
        pattern = Perm.to_standard(self.pattern[index] for index in indices)
        vertical = [0]
        vertical.extend(index + 1 for index in indices)
        vertical.append(len(self) + 1)
        horizontal = [0]
        horizontal.extend(sorted(self.pattern[index] + 1 for index in indices))
        horizontal.append(len(self) + 1)
        shading = frozenset(
            (x, y)
            for x in range(len(pattern) + 1)
            for y in range(len(pattern) + 1)
            if (
                self.is_shaded(
                    (vertical[x], horizontal[y]),
                    (vertical[x + 1] - 1, horizontal[y + 1] - 1),
                )
                and self.is_pointfree(
                    (vertical[x], horizontal[y]),
                    (vertical[x + 1] - 1, horizontal[y + 1] - 1),
                )
            )
        )
        return MeshPatt(pattern, shading)

    def flip_horizontal(self):
        """Return self flipped horizontally which is equivalent to the
        complement.

        Returns: <permuta.MeshPatt>
            The complement of the meshpatt.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1)})).flip_horizontal()
            MeshPatt(Perm((0,)), [(0, 0)])
            >>> MeshPatt(Perm((0, 2, 1)),
            ... frozenset({(0, 1), (0, 2), (0, 3)})).flip_horizontal()
            MeshPatt(Perm((2, 0, 1)), [(0, 0), (0, 1), (0, 2)])
        """
        return self.complement()

    def flip_vertical(self):
        """Return self flipped vertically which is equivalent to the
        meshpatt reversed.

        Returns: <permuta.MeshPatt>
            The meshpatt reversed.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1)})).flip_vertical()
            MeshPatt(Perm((0,)), [(1, 1)])
            >>> MeshPatt(Perm((2, 1, 0)),
            ... frozenset({(3, 2), (3, 3), (0, 2)})).flip_vertical()
            MeshPatt(Perm((0, 1, 2)), [(0, 2), (0, 3), (3, 2)])
        """
        return self.reverse()

    def flip_diagonal(self):
        """Return self flipped along the diagonal which is equivalent to the
        'inverse' of the pattern.

        Returns: <permuta.MeshPatt>
            The 'inverse' of the meshpatt.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1)})).inverse()
            MeshPatt(Perm((0,)), [(1, 0)])
        """
        return self.inverse()

    def _rotate_right(self):
        """Return the pattern rotated 90 degrees to the right.

        Returns: <permuta.MeshPatt>
            The meshpatt rotated 90 degrees to the right.

        Examples:
            >>> MeshPatt(Perm((0,)),
            ... frozenset({(0, 1), (1, 1)}))._rotate_right()
            MeshPatt(Perm((0,)), [(1, 0), (1, 1)])
        """
        return MeshPatt(
            self.pattern.rotate(),
            set(
                [
                    _rotate_right(len(self.pattern), coordinate)
                    for coordinate in self.shading
                ]
            ),
        )

    def _rotate_left(self):
        """Return the pattern rotated 90 degrees to the left.

        Returns: <permuta.MeshPatt>
            The meshpatt rotated 90 degrees to the left.

        Examples:
            >>> MeshPatt(Perm((0,)),
            ... frozenset({(0, 1), (1, 1)}))._rotate_left()
            MeshPatt(Perm((0,)), [(0, 0), (0, 1)])
        """
        return MeshPatt(
            self.pattern.rotate(3),
            set(
                [
                    _rotate_left(len(self.pattern), coordinate)
                    for coordinate in self.shading
                ]
            ),
        )

    def _rotate_180(self):
        """Return the pattern rotated 180 degrees.

        Returns: <permuta.MeshPatt>
            The meshpatt rotated 180 degrees.

        Examples:
            >>> MeshPatt(Perm((0,)), frozenset({(0, 1), (1, 1)}))._rotate_180()
            MeshPatt(Perm((0,)), [(0, 0), (1, 0)])
        """
        return MeshPatt(
            self.pattern.rotate(2),
            set(
                [
                    _rotate_180(len(self.pattern), coordinate)
                    for coordinate in self.shading
                ]
            ),
        )

    def all_symmetries(self):
        """Return the set of all symmetries of the mesh pattern.

        Returns: <set>
            All the symmetries of a mesh pattern.
        """
        symmetries = set()
        current = self
        symmetries.add(current)
        symmetries.add(current.inverse())
        for i in range(3):
            current = current._rotate_left()
            symmetries.add(current)
            symmetries.add(current.inverse())
        return symmetries

    def shade(self, positions):
        """Returns the mesh pattern with the added shadings given by positions.

        Args:
            positions: tuple or <collections.Iterable>
                The shading given as a single coordinate or as an iterable of
                coordinates.

        Raises:
            ValueError:
                Bad argument, but correct type.

        Returns: <permuta.MeshPatt>
            The new meshpatt with the added shadings.
        """
        if isinstance(positions, tuple):
            if len(positions) == 0:
                raise ValueError(
                    "Element is not a valid shading coordinate: '{}'".format(positions)
                )
            if isinstance(positions[0], numbers.Integral):
                positions = set([positions])
        positions = set(positions)

        return MeshPatt(self.pattern, self.shading | positions)

    def add_point(self, pos, shade_dir=DIR_NONE, safe=True):
        """Returns a mesh pattern with a point added in the box at position
        pos.  If shade_dir is specified adds shading in that direction.

        Args:
            pos: tuple
                Coordinates corresponding to a box in the pattern.
            shade_dir: int
                The direction which to shade within the box.
            safe: bool, optional
                True to check if box must not be shaded.

        Raises:
            ValueError:
                Bad box argument, if it is already shaded.
            TypeError:
                Bad argument type.

        Returns: <permuta.MeshPatt>
            The mesh pattern with a point added in pos.
        """
        x, y = pos
        if not isinstance(x, numbers.Integral) or not isinstance(y, numbers.Integral):
            message = "Element is not a tuple of integers: '{}'".format(pos)
            raise TypeError(message)
        if safe and (x, y) in self.shading:
            message = "Can not add point to shaded pos: '{}'"
            raise ValueError(message)

        perm = [v if v < y else (v + 1) for v in self.pattern]
        nperm = perm[:x] + [y] + perm[x:]
        nshading = set()
        for (a, b) in self.shading:
            if a < x:
                nx = [a]
            elif a == x:
                nx = [a, a + 1]
            else:
                nx = [a + 1]

            if b < y:
                ny = [b]
            elif b == y:
                ny = [b, b + 1]
            else:
                ny = [b + 1]

            for na in nx:
                for nb in ny:
                    nshading.add((na, nb))

        if shade_dir == DIR_EAST:
            nshading.add((x + 1, y))
            nshading.add((x + 1, y + 1))
        elif shade_dir == DIR_NORTH:
            nshading.add((x, y + 1))
            nshading.add((x + 1, y + 1))
        elif shade_dir == DIR_WEST:
            nshading.add((x, y))
            nshading.add((x, y + 1))
        elif shade_dir == DIR_SOUTH:
            nshading.add((x, y))
            nshading.add((x + 1, y))

        return MeshPatt(Perm(nperm), nshading)

    def add_increase(self, pos):
        """Adds an increasing pattern (0, 1) into the given coordinate.

        Args:
            pos: tuple
                The coordinate of the position to insert (0, 1) into.

        Returns: <permuta.MeshPatt>
            The new pattern with the increased pattern inserted into pos.

        Examples:
            >>> MeshPatt((0,)).add_increase((0, 0))
            MeshPatt(Perm((0, 1, 2)), [])
        """
        x, y = pos
        return self.add_point((x, y)).add_point((x + 1, y + 1))

    def add_decrease(self, pos):
        """Adds an decreasing pattern (1, 0) into the given coordinate.

        Args:
            pos: tuple
                The coordinate of the position to insert (1, 0) into.

        Returns: <permuta.MeshPatt>
            The new pattern with the decreasing pattern inserted into pos.

        Examples:
            >>> MeshPatt((0,)).add_decrease((0, 0))
            MeshPatt(Perm((1, 0, 2)), [])
        """
        x, y = pos
        return self.add_point((x, y)).add_point((x + 1, y))

    #
    # Occurrence/Avoidance/Containment methods
    #

    def contains(self, *patts):
        """Check if self contains patts.

        Args:
            self:
                A MeshPatt.
            patts: <permuta.Patt> argument list
                Classical/mesh patterns.

        Returns: <bool>
            True if and only if all patterns in patts are contained in self.
        """
        return all(patt in self for patt in patts)

    def avoids(self, *patts):
        """Check if self avoids patts.

        Args:
            self:
                A MeshPatt.
            patts: <permuta.Patt> argument list
                Classical/mesh patterns.

        Returns: <bool>
            True if and only if self avoids all patterns in patts.
        """
        return all(patt not in self for patt in patts)

    def occurrences_in(self, patt):
        """
        Find all indices of self in patt.

        Args:
            self:
                The mesh pattern whose occurrences are to be found.
            patt: <permuta.Perm> or <permuta.MeshPatt>
                The patt to search for occurrences in.

        Yields: numbers.Integral
            The indices of the occurrences of self in perm. Each yielded
            element l is a list of integer indices of the pattern `patt` such
            that:

            If `patt` is a classical pattern:
                self.pattern == permuta.Perm.to_standard([patt[i] for i in l])
                and that no element not in the occurrence falls within
                self.shading.

            If `patt` is a mesh pattern:
                The list of indices describe a classical occurrence of
                `patt.patt` as described above and the shading of the submesh
                pattern induced by this occurrence is a superset of shading of
                `self.shading`.
        """
        if isinstance(patt, Perm):
            for candidate_indices in self.pattern.occurrences_in(patt):
                candidate = [patt[index] for index in candidate_indices]
                x = 0
                for element in patt:
                    if element in candidate:
                        x += 1
                        continue
                    y = sum(
                        1
                        for candidate_element in candidate
                        if candidate_element < element
                    )
                    if (x, y) in self.shading:
                        break
                else:
                    # No unused point fell within shading
                    yield list(candidate_indices)
        elif isinstance(patt, MeshPatt):
            for occurrence in self.occurrences_in(patt.pattern):
                candidate_sub_mesh_patt = patt.sub_mesh_pattern(occurrence)
                if set(self.shading) <= set(candidate_sub_mesh_patt.shading):
                    yield list(occurrence)
        else:
            raise ValueError("Variable 'patt' needs to be either a Perm or " "MeshPatt")

    #
    # Other methods
    #

    def is_shaded(self, lower_left, upper_right=None):
        """Check if a region of the grid is shaded.

        Args:
            self:
                A mesh pattern.
            lower_left: (int, int)
                A shading coordinate of self.
            upper_right: (int, int)
                A shading coordinate of self.

        Raises:
            ValueError:
                Bad argument, but correct type.

        Returns: bool
            If upper_right is None, then True if and only if lower_left is
            shaded; otherwise, True if and only if all regions (x, y) for x in
            the range lower_left[0] to upper_right[0] (inclusive) and for y in
            the range lower_left[1] to upper_right[1] (inclusive) are shaded.
        """
        if (lower_left[0] < 0 or lower_left[1] < 0) or (
            lower_left[0] > len(self) or lower_left[1] > len(self)
        ):
            message = "Element out of range: '{}'".format(lower_left)
            raise ValueError(message)
        elif upper_right is None:
            return lower_left in self.shading
        elif (upper_right[0] < 0 or upper_right[1] < 0) or (
            upper_right[0] > len(self) or upper_right[1] > len(self)
        ):
            message = "Element out of range: '{}'".format(upper_right)
            raise ValueError(message)
        elif lower_left[0] > upper_right[0] or lower_left[1] > upper_right[1]:
            message = (
                "Elements do not correspond to lower left and upper"
                " right of a non-empty rectangle: '{}' '{}'"
            ).format(lower_left, upper_right)
            raise ValueError(message)
        else:
            left, lower = lower_left
            right, upper = upper_right
            for x in range(left, right + 1):
                for y in range(lower, upper + 1):
                    if (x, y) not in self.shading:
                        return False
            return True

    def is_pointfree(self, lower_left, upper_right):
        """Check if a region in the grid has no points.

        Args:
            self:
                A mesh pattern.
            lower_left: (int, int)
                A point coordinate of self.
            upper_right: (int, int)
                A point coordinate of self.

        Raises:
            ValueError:
                Bad argument, but correct type.

        Returns: bool
            True if and only if all points self[x] for x in the range
            lower_left[0] + 1 to upper_right[0] - 1 (inclusive) have values
            less than lower_left[1] or greater than or equal to upper_right[1].
        """
        if (lower_left[0] < 0 or lower_left[1] < 0) or (
            lower_left[0] > len(self) or lower_left[1] > len(self)
        ):
            message = "Element out of range: '{}'".format(lower_left)
            raise ValueError(message)
        elif (upper_right[0] < 0 or upper_right[1] < 0) or (
            upper_right[0] > len(self) or upper_right[1] > len(self)
        ):
            message = "Element out of range: '{}'".format(upper_right)
            raise ValueError(message)
        elif lower_left[0] > upper_right[0] or lower_left[1] > upper_right[1]:
            message = (
                "Elements do not correspond to lower left and upper "
                "right of a non-empty rectangle: '{}' '{}'"
            ).format(lower_left, upper_right)
            raise ValueError(message)
        else:
            left, lower = lower_left
            right, upper = upper_right
            for x in range(left, right):
                if lower <= self.pattern[x] < upper:
                    return False
        return True

    def _can_shade(self, pos):
        """Checks if the box at pos can be shaded according to the Shading
        Lemma(northeast).

        Args:
            pos: tuple
                The coordinates of the box to check.

        Returns: tuple
            The point that is 'moved' to shade the box at pos.
        """
        i, j = pos
        # if pos is shaded
        if (i, j) in self.shading:
            return False
        # if pos does not have point in lower-left
        if i - 1 < 0 or self.pattern[i - 1] != j - 1:
            return False
        # if box in south-west direction is shaded
        if (i - 1, j - 1) in self.shading:
            return False
        c = 0
        # only one of the boxes to the left and down can be shaded
        if (i, j - 1) in self.shading:
            c += 1
        if (i - 1, j) in self.shading:
            c += 1
        if c == 2:
            return False

        # if the box on the lower side of the horizontal line is shaded then
        # the upper one must be shaded
        for k in range(len(self.pattern) + 1):
            if k == i - 1 or k == i:
                continue
            if (k, j - 1) in self.shading and (k, j) not in self.shading:
                return False
        # if the box on the left side of the vertical line is shaded then the
        # right one must be shaded
        for k in range(len(self.pattern) + 1):
            if k == j - 1 or k == j:
                continue
            if (i - 1, k) in self.shading and (i, k) not in self.shading:
                return False
        return (i - 1, j - 1)

    def can_shade(self, pos):
        """Returns whether it is possible to shade the box at position pos
        according to the Shading Lemma. Every direction is checked and the list
        of the values of the adjacent points to the box that can be used is
        returned.

        Args:
            pos: tuple
                The position of the box to check.

        Returns: list
            The values of the points in the permutation adjacent to the box at
            pos that can be used to shade the box.

        Examples:
            >>> MeshPatt((0,),[(0, 0)]).can_shade((1, 1))
            []
            >>> MeshPatt((1, 2, 0),
            ... [(2, 2),(3, 0),(3, 2),(3, 3)]).can_shade((1, 2))
            [1, 2]
        """
        mp = self
        poss = []
        for i in range(4):
            ans = mp._can_shade(pos)
            if ans:
                for j in range((-i) % 4):
                    ans = _rotate_right(len(self.pattern) - 1, ans)
                poss.append(ans[1])
            mp = mp.rotate_right()
            pos = _rotate_right(len(self.pattern), pos)
        return poss

    def _can_simul_shade(self, pos1, pos2):
        if pos1[1] < pos2[1]:
            pos1, pos2 = pos2, pos1
        # There must be a point at (pos1[0]-1, pos1[1]-1)
        if pos1[0] == 0 or self.pattern[pos1[0] - 1] != pos1[1] - 1:
            return False
        if pos1[0] != pos2[0] or pos1[1] - 1 != pos2[1]:
            # the pos1 must be directly above pos2
            return False

        if pos1 in self.shading:
            return False
        if pos2 in self.shading:
            return False
        # The boxes surrounding the point (pos1[0]-1, pos1[1] - 1) must be
        # empty
        if (pos1[0] - 1, pos1[1]) in self.shading:
            return False
        if (pos2[0] - 1, pos2[1]) in self.shading:
            return False

        # Check the boxes on each side of the vertical line of pos1[0], if the
        # box on the left side is shaded then the box on the right side must be
        # shaded.
        for y in range(len(self.pattern) + 1):
            if y == pos1[1] or y == pos1[1] - 1:
                continue
            if (pos1[0] - 1, y) in self.shading and (pos1[0], y) not in self.shading:
                return False

        # Check the boxes on each side of the horizontal line of pos1[1]-1,
        # they must match.
        for x in range(len(self.pattern) + 1):
            if x == pos1[0] or x == pos1[0] - 1:
                continue
            if ((x, pos1[1]) in self.shading) != ((x, pos2[1]) in self.shading):
                return False
        return (pos1[0] - 1, pos1[1] - 1)

    def can_simul_shade(self, pos1, pos2):
        """Returns whether it is possible to shade the boxes at positions pos1,
        pos2 according to the Shading Lemma. Every direction is checked and the
        list of the values of the adjacent points to the box that can be used
        is returned.

        Args:
            pos1: tuple
                The position of the first box to check.
            pos2: tuple
                The position of the second box to check.

        Returns: list
            The values of the points in the permutation adjacent to the boxes
            at pos1 and pos2 that can be used to shade the box.

        Examples:

        """
        mp = self
        poss = []
        for i in range(4):
            ans = mp._can_simul_shade(pos1, pos2)
            if ans:
                for j in range((-i) % 4):
                    ans = _rotate_right(len(self.pattern) - 1, ans)
                poss.append(ans[1])
            mp = mp.rotate_right()
            pos1 = _rotate_right(len(self.pattern), pos1)
            pos2 = _rotate_right(len(self.pattern), pos2)
        return poss

    can_shade2 = can_simul_shade

    def shadable_boxes(self):
        """Returns a dictionary of all tuples of shadable boxes with the
        shading lemma, with the keys as the points used to shade the tuples of
        boxes.

        Returns: dict
            Dictionary with keys as points and values as tuples of boxes.

        Examples:
        >>> sh = {(3, 2), (3, 0), (4, 2), (1, 0), (0, 3), (1, 2), (0, 4),
        ... (0, 2)}
        >>> m = MeshPatt(Perm((0, 3, 1, 2)), sh)
        >>> dict(m.shadable_boxes())
        {0: [((0, 0),)], 3: [((1, 3),), ((1, 3), (1, 4)), ((1, 4),)]}
        """
        shadable = collections.defaultdict(list)
        for i in range(len(self) + 1):
            for j in range(len(self) + 1):
                points = self.can_shade((i, j))
                for p in points:
                    shadable[p].append(((i, j),))
                if i < len(self):
                    points = self.can_simul_shade((i, j), (i + 1, j))
                    for p in points:
                        shadable[p].append(((i, j), (i + 1, j)))
                if j < len(self):
                    points = self.can_simul_shade((i, j), (i, j + 1))
                    for p in points:
                        shadable[p].append(((i, j), (i, j + 1)))
        return shadable

    def non_pointless_boxes(self):
        """ Returns the coordinates of the boxes that have a point on their
        boundaries in one of their four corners.

        Returns: set
            The set of boxes with points in one of the corners.

        Examples:
            >>> m = MeshPatt(Perm((0, 1)), frozenset({(0, 1), (2, 0), (0, 2)}))
            >>> sorted(m.non_pointless_boxes())
            [(0, 0), (0, 1), (1, 0), (1, 1), (1, 2), (2, 1), (2, 2)]
        """
        res = []
        for i, v in enumerate(self.pattern):
            res.extend([(i + 1, v + 1), (i, v + 1), (i, v), (i + 1, v)])
        return set(res)

    def has_anchored_point(self):
        """Checks if the mesh pattern has any point anchored to the boundary.
        Returns a tuple (right, top, left, bottom) where each value represent
        whether the point is anchored to the corresponding direction.

        Returns: tuple
            Each boolean in the tuple tells whether the mesh pattern is
            anchored to the corresponding direction.

        Examples:
            >>> m = MeshPatt(Perm((0, 1)), {(0, 0), (1, 0), (2, 0), (1, 1)})
            >>> m.has_anchored_point()
            (False, False, False, True)
        """
        right = all((len(self), i) in self.shading for i in range(len(self) + 1))
        top = all((i, len(self)) in self.shading for i in range(len(self) + 1))
        left = all((0, i) in self.shading for i in range(len(self) + 1))
        bottom = all((i, 0) in self.shading for i in range(len(self) + 1))
        return (right, top, left, bottom)

    def rank(self):
        """Computes the rank of the mesh pattern, the bit string of the
        shadings interpreted as an integer.

        Returns: numbers.Integral
            The rank of the mesh pattern.

        Examples:
            >>> sh = {(0, 0), (3, 0), (0, 2), (2, 1), (2, 3), (1, 2), (3, 3),
            ... (3, 1), (1, 1)}
            >>> m = MeshPatt(Perm((1, 0, 2)), sh)
            >>> rank = m.rank()
            >>> rank
            47717
            >>> bin(rank)
            '0b1011101001100101'
        """
        res = 0
        for (x, y) in self.shading:
            res |= 1 << (x * (len(self.pattern) + 1) + y)
        return res

    def ascii_plot(self, cell_size=1):
        """Return an ascii plot of the given Permutation.

        Args:
            self:
                A perm.
            cell_size: <int>
                The size of the cell of the grid

        Returns: <str>
            The ascii art string of the permutation

        Examples:
            >>> print(MeshPatt((0,1,2), [(2,1), (3,0), (3,1), (3,2), (3,3)]
            ...     ).ascii_plot())
             | | |▒
            -+-+-●-
             | | |▒
            -+-●-+-
             | |▒|▒
            -●-+-+-
             | | |▒
        """

        def roundrobin(*iterables):
            "roundrobin('ABC', 'D', 'EF') --> A D E B F C"
            # Recipe credited to George Sakkis
            num_active = len(iterables)
            nexts = cycle(iter(it).__next__ for it in iterables)
            while num_active:
                try:
                    for next in nexts:
                        yield next()
                except StopIteration:
                    # Remove the iterator we just exhausted from the cycle.
                    num_active -= 1
                    nexts = cycle(islice(nexts, num_active))

        def fill_char(c):
            shading_char = "\u2592"
            if c in self.shading:
                return shading_char
            elif c[0] == len(self):
                return ""
            else:
                return " "

        if cell_size < 1:
            raise ValueError("`cell_size` must be positive")
        empty_char = "+"
        point_char = "\u25cf"
        n = self.pattern.__len__()
        array = [[empty_char for i in range(n)] for j in range(n)]
        for i in range(n):
            array[self.pattern[i]][i] = point_char
        array.reverse()
        lines = [("-" * cell_size).join([""] + line + [""]) + "\n" for line in array]
        vlines = [
            ("|".join(fill_char((j, i)) * cell_size for j in range(n + 1)) + "\n")
            * cell_size
            for i in range(n + 1)
        ]
        vlines.reverse()
        s = "".join(roundrobin(vlines, lines))
        return s[:-1]

    def to_tikz(self):
        """
        Return the tikz code to plot the mesh pattern. The tikz code requires
        the TikZ library patter.

        Returns: str
            The LaTeX code for the TikZ figure of the pattern.
        """
        s = r"\begin{tikzpicture}"
        s += r"[scale=.3,baseline=(current bounding box.center)]"
        s += "\n\t"
        s += r"\foreach \x in {1,...," + str(len(self)) + "} {"
        s += "\n\t\t"
        s += r"\draw[ultra thin] (\x,0)--(\x," + str(len(self) + 1) + "); %vline"
        s += "\n\t\t"
        s += r"\draw[ultra thin] (0,\x)--(" + str(len(self) + 1) + r",\x); %hline"
        s += 2 * "\n\t"
        s += r"}"
        print(self.shading)
        for cell in sorted(self.shading):
            s += "\n\t"
            s += r"\fill[pattern color = black!75, pattern=north east lines] "
            s += str(cell) + r" rectangle +(1,1);"
            print(cell)
        for (i, e) in enumerate(self.pattern):
            s += "\n\t"
            s += (
                r"\draw[fill=black] ("
                + str(i + 1)
                + ","
                + str(e + 1)
                + ") circle (5pt);"
            )
        s += "\n"
        s += r"\end{tikzpicture}"
        return s

    #
    # Static methods
    #

    @staticmethod
    def unrank(pattern, number):
        """Return the number-th shading of pattern.

        Args:
            pattern: <permuta.Perm> or <collections.Iterable>
                An perm or an iterable corresponding to a legal perm.
            number: <numbers.Integral>
                An integer of which binary representation corresponds to a
                legal shading.
        Raises:
            TypeError:
                Bad argument type.
            ValueError:
                Bad argument, but correct type.
        Examples:
            >>> bin(22563)
            '0b101100000100011'
            >>> MeshPatt.unrank((0, 1, 2), 386)
            MeshPatt(Perm((0, 1, 2)), [(0, 1), (1, 3), (2, 0)])
        """
        if not isinstance(number, numbers.Integral):
            message = "'{}' object is not an integer".format(repr(number))
            raise TypeError(message)
        if not (0 <= number < 2 ** ((len(pattern) + 1) ** 2)):
            message = "Element out of range: '{}'".format(number)
            raise ValueError(message)
        bound = len(pattern) + 1
        shading = set(
            (index // bound, index % bound)
            for index, bit in enumerate(reversed(bin(number)[2:]))
            if bit == "1"
        )
        return MeshPatt(pattern, shading)

    @staticmethod
    def random(length):
        """Return a random mesh pattern of the specified length.

        Args:
            length: <numbers.Integral>
                The length of the random pattern.

        Examples:
            >>> mp = set(MeshPatt.unrank(Perm((0, )), i) for i in range(0, 16))
            >>> MeshPatt.random(1) in mp
            True
            >>> len(MeshPatt.random(4))
            4
        """
        return MeshPatt.unrank(
            Perm.random(length), random.randint(0, 2 ** ((length + 1) ** 2) - 1)
        )

    #
    # Dunder methods
    #

    def __repr__(self):  # pragma: no cover
        return "MeshPatt({}, {})".format(repr(self.pattern), sorted(self.shading))

    def __str__(self):
        return "({}, {})".format(self.pattern, sorted(self.shading))

    def __len__(self):
        return len(self.pattern)

    def __bool__(self):
        return bool(self.pattern) or bool(self.shading)

    def __lt__(self, other):
        return (self.pattern, sorted(self.shading)) < (
            other.pattern,
            sorted(other.shading),
        )

    def __le__(self, other):
        return (self.pattern, sorted(self.shading)) <= (
            other.pattern,
            sorted(other.shading),
        )

    def __gt__(self, other):
        return other.__lt__(self)

    def __ge__(self, other):
        return other.__le__(self)

    def __contains__(self, patt):
        """Check if self contains patt.


        Args:
            self:
                A perm.
            patt: <permuta.Patt>
                A classical/mesh pattern.

        Returns: <bool>
            True if and only if the pattern patt is contained in self.
        """
        return any(True for _ in patt.occurrences_in(self))


def _rotate_right(length, element):
    """Rotate an element of the Cartesian product of {0,...,length} clockwise.

    Args:
        length: <numbers.Integral>
            The size of the area(of mesh).
        element: tuple
            A cartesiean coordinate within [0,...,length]x[0,...,length]

    Returns: tuple
        The input point rotated within the box of size length x length.
    """
    x, y = element
    return (y, length - x)


def _rotate_left(length, element):
    """Rotate an element of the Cartesian product of {0,...,length}
    counterclockwise.

    Args:
        length: <numbers.Integral>
            The size of the area(of mesh).
        element: tuple
            A cartesiean coordinate within [0,...,length]x[0,...,length]

    Returns: tuple
        The input point rotated within the box of size length x length.
    """
    x, y = element
    return (length - y, x)


def _rotate_180(length, element):
    """Rotate an element of the Cartesian product of {0,...,length}
    180-degrees.

    Args:
        length: <numbers.Integral>
            The size of the area(of mesh).
        element: tuple
            A cartesiean coordinate within [0,...,length]x[0,...,length]

    Returns: tuple
        The input point rotated within the box of size length x length.
    """
    x, y = element
    return (length - x, length - y)



# TODO: Module docstring

import abc
from collections.abc import Iterable

from ..meshpatt import MeshPatt
from ..perm import Perm
from .descriptor import Descriptor


class AbstractBasis(Descriptor, tuple, abc.ABC):
    @property
    @abc.abstractmethod
    def ALLOWED_BASIS_ELEMENT_TYPES(self):
        raise NotImplementedError

    def __new__(cls, patts):
        return tuple.__new__(cls).union(patts, cls.ALLOWED_BASIS_ELEMENT_TYPES)

    def union(self, patts, patt_types):
        if not isinstance(patts, Iterable):
            raise TypeError("Non-iterable argument cannot be unified with basis")

        # Input cleaning
        patts = set([patts] if isinstance(patts, patt_types) else patts)

        if not patts:
            # Empty set of patts added to basis
            return self

        # Make sure the elements are patterns
        for patt in patts:
            if not isinstance(patt, patt_types):
                raise TypeError(
                    "Elements of a basis should all be of type(s) {}".format(patt_types)
                )

        # Add basis patts and sort
        patts.update(self)
        patts = sorted(patts)  # Necessarily non-empty

        # The new basis
        new_basis = []

        # The list of basis patts used for the new basis
        basis_patts_used = []

        # The list of new patts used for the new basis
        new_patts_used = []

        if patts[0] in [c() for c in patt_types]:
            new_basis.append(patts[0])
            new_patts_used.append(patts[0])
        else:
            patts_iter = iter(patts)
            basis_iter = iter(self)
            for basis_perm in basis_iter:
                # Add perms up to and including this basis perm to the basis
                while True:
                    patt = next(patts_iter)
                    if patt == basis_perm:
                        # Add if it avoids the new perms used
                        if patt.avoids(*new_patts_used):
                            new_basis.append(patt)
                            basis_patts_used.append(patt)
                        break
                    elif patt.avoids(*new_basis):
                        # Add it if it avoids the new basis perms
                        new_basis.append(patt)
                        new_patts_used.append(patt)
            for patt in patts_iter:
                # All perms left over weren't in the basis before
                if patt.avoids(*new_basis):
                    new_basis.append(patt)
                    new_patts_used.append(patt)

        # Return either the unmodified basis or a new basis
        if new_patts_used:
            return tuple.__new__(self.__class__, new_basis)
        else:
            return self

    def is_polynomial(self):
        return True  # TODO

    def __eq__(self, other):
        return isinstance(other, self.__class__) and tuple.__eq__(self, other)

    def __hash__(self):
        return tuple.__hash__(self)

    def __repr__(self):
        return "{}({})".format(self.__class__.__qualname__, tuple.__repr__(self))

    def __str__(self):
        return "{{{}}}".format(", ".join(str(p) for p in self))


class Basis(AbstractBasis):
    """A basis class.

    A PermSet can be built with a Basis instance by using the basis provided
    to it to see if a perm should be in the PermSet or not. Additionally,
    various fast methods exist to build a PermSet defined by a basis.
    """

    ALLOWED_BASIS_ELEMENT_TYPES = (Perm,)


class MeshBasis(AbstractBasis):
    ALLOWED_BASIS_ELEMENT_TYPES = (MeshPatt,)

    def __new__(cls, patts):
        return super().__new__(
            cls,
            {
                patt if isinstance(patt, MeshPatt) else MeshPatt(patt, [])
                for patt in patts
            },
        )


def detect_basis_cls(basis):
    # Argument can be the actual basis class
    if basis in AbstractBasis.__subclasses__():
        return basis

    # Argument can be an instance of the basis classes
    for BasisCls in AbstractBasis.__subclasses__():
        if isinstance(basis, BasisCls):
            return BasisCls

    # Argument can be an object or an iterable of objects that makes up a basis
    if isinstance(basis, Perm) or all(isinstance(patt, Perm) for patt in basis):
        return Basis
    elif isinstance(basis, MeshPatt) or all(
        isinstance(patt, (Perm, MeshPatt)) for patt in basis
    ):
        return MeshBasis
    else:
        raise ValueError("A basis can only contain Perms and MeshPatts.")



import random

import pytest
from permuta import Perm, PermSet


def test_init():
    with pytest.raises(ValueError):
        Perm([0, 1, 1], check=True)
    with pytest.raises(ValueError):
        Perm([1, 0, 1], check=True)
    with pytest.raises(ValueError):
        Perm([0, 0], check=True)
    with pytest.raises(ValueError):
        Perm([1], check=True)
    with pytest.raises(ValueError):
        Perm((1,), check=True)
    with pytest.raises(TypeError):
        Perm(101, check=True)
    with pytest.raises(TypeError):
        Perm(-234, check=True)
    with pytest.raises(TypeError):
        Perm(None, check=True)
    with pytest.raises(TypeError):
        Perm([0.1, 0.2, 0.3], check=True)
    Perm(check=True)
    Perm([], check=True)
    Perm((0,), check=True)
    Perm([0], check=True)
    Perm([3, 0, 2, 1], check=True)
    Perm(set([0, 1, 2]), check=True)


def test_to_standard():
    def gen(perm):
        res = list(perm)
        add = 0
        for i in perm.inverse():
            add += random.randint(0, 10)
            res[i] += add
        return Perm(res)

    perm = Perm.to_standard(x for x in range(10))
    Perm.to_standard(x for x in range(3, 10))

    for _ in range(100):
        perm = Perm.random(random.randint(0, 20))
        assert perm == Perm.to_standard(perm)
        assert perm == Perm.to_standard(gen(perm))


def test_from_string():
    assert Perm.from_string("203451") == Perm((2, 0, 3, 4, 5, 1))
    assert Perm.from_string("40132") == Perm((4, 0, 1, 3, 2))
    assert Perm.from_string("0") == Perm((0,))

    for _ in range(100):
        perm = Perm.random(random.randint(0, 10))
        assert perm == Perm.from_string("".join(map(str, perm)))


def test_str_representation():
    assert str(Perm((0, 2, 1))) == "021"
    assert (
        str(Perm((0, 11, 1, 10, 2, 9, 3, 8, 4, 7, 5, 6)))
        == "(0)(11)(1)(10)(2)(9)(3)(8)(4)(7)(5)(6)"
    )


def test_one_based():
    assert Perm.one_based((4, 1, 3, 2)) == Perm((3, 0, 2, 1))
    assert Perm.one_based((1,)) == Perm((0,))
    assert Perm.one_based(()) == Perm()

    for _ in range(100):
        p = list(range(1, random.randint(2, 20)))
        random.shuffle(p)
        assert Perm.one_based(p) == Perm([i - 1 for i in p])


def test_identity():
    for length in range(11):
        assert Perm.identity(length) == Perm(range(length))


def test_random():
    for length in range(11):
        for _ in range(10):
            perm = Perm.random(length)
            assert len(perm) == length
            Perm(perm)


def test_monotone_increasing():
    for length in range(11):
        assert Perm.monotone_increasing(length) == Perm(range(length))


def test_monotone_decreasing():
    for length in range(11):
        assert Perm.monotone_decreasing(length) == Perm(range(length - 1, -1, -1))


def test_unrank():
    assert Perm.unrank(0) == Perm()
    assert Perm.unrank(1) == Perm((0,))
    assert Perm.unrank(2) == Perm((0, 1))
    assert Perm.unrank(3) == Perm((1, 0))
    assert Perm.unrank(4) == Perm((0, 1, 2))
    assert Perm.unrank(5) == Perm((0, 2, 1))
    assert Perm.unrank(6) == Perm((1, 0, 2))
    assert Perm.unrank(10) == Perm((0, 1, 2, 3))
    amount = 1 + 1 + 2 + 6 + 24
    assert Perm.unrank(amount) == Perm((0, 1, 2, 3, 4))
    amount = (1 + 1 + 2 + 6 + 24 + 120) - 1
    assert Perm.unrank(amount) == Perm((4, 3, 2, 1, 0))
    with pytest.raises(AssertionError):
        Perm.unrank(-1)
    with pytest.raises(AssertionError):
        Perm.unrank(6, 3)


def test_unrank_2():
    length = 7
    for number, perm in enumerate(PermSet(length)):
        assert Perm.unrank(number, length) == perm


def test_contained_in():
    def generate_contained(n, perm):
        for i in range(len(perm), n):
            r = random.randint(1, len(perm) + 1)
            for i in range(len(perm)):
                if perm[i] >= r:
                    perm[i] += 1
            x = random.randint(0, len(perm))
            perm = perm[:x] + [r] + perm[x:]
        return Perm(perm)

    assert Perm([3, 7, 0, 8, 1, 6, 5, 2, 4]).contained_in(
        Perm([3, 7, 0, 8, 1, 6, 5, 2, 4])
    )
    assert Perm([3, 7, 0, 8, 1, 6, 5, 2, 4]).contains(Perm([3, 7, 0, 8, 1, 6, 5, 2, 4]))
    assert Perm([]).contained_in(Perm([]))
    assert Perm([]).contained_in(Perm([3, 7, 0, 8, 1, 6, 5, 2, 4]))
    assert Perm([0]).contained_in(Perm([0]))
    assert not Perm([7, 3, 0, 8, 1, 6, 5, 2, 4]).contained_in(
        Perm([3, 7, 0, 8, 1, 6, 5, 2, 4])
    )

    for i in range(100):
        n = random.randint(0, 4)
        patt = PermSet(n).random()
        perm = generate_contained(random.randint(n, 8), list(patt))
        assert patt.contained_in(perm)
        assert perm.contains(patt)

    assert not Perm([0]).contained_in(Perm([]))
    assert not Perm([0, 1]).contained_in(Perm([]))
    assert not Perm([0, 1]).contained_in(Perm([0]))
    assert not Perm([1, 0]).contained_in(Perm([0, 1]))
    assert not Perm([0, 1, 2]).contained_in(Perm([0, 1]))
    assert not Perm([1, 0, 2]).contained_in(Perm([0, 1, 3, 4, 2]))
    assert not Perm([0, 1, 2]).contained_in(Perm([2, 1, 3, 0]))
    assert not Perm([2, 1, 3, 0]).contained_in(Perm([2, 0, 3, 1]))
    assert not Perm([0, 2, 1]).contained_in(Perm([2, 0, 1, 3]))
    assert not Perm([2, 0, 1, 3]).contained_in(Perm([5, 3, 2, 7, 1, 0, 6, 4]))
    assert not Perm([0, 1, 2, 3]).contained_in(Perm([4, 7, 5, 1, 6, 2, 3, 0]))


def test_count_occurrences_in():
    assert Perm([]).count_occurrences_in(Perm([4, 1, 2, 3, 0])) == 1
    assert Perm([0]).count_occurrences_in(Perm([4, 1, 2, 3, 0])) == 5
    assert Perm([0, 1]).count_occurrences_in(Perm([4, 1, 2, 3, 0])) == 3
    assert Perm([1, 0]).count_occurrences_in(Perm([4, 1, 2, 3, 0])) == 7
    assert Perm([4, 1, 2, 3, 0]).count_occurrences_in(Perm([])) == 0
    assert Perm([4, 1, 2, 3, 0]).count_occurrences_in(Perm([1, 0])) == 0


def test_count_occurrences_of():
    assert Perm([4, 1, 2, 3, 0]).count_occurrences_of(Perm([1, 0])) == 7
    assert Perm([4, 1, 2, 3, 0]).count_occurrences_of(Perm([0, 1])) == 3
    assert Perm((1, 3, 4, 0, 2, 5)).count_occurrences_of(Perm((0, 2, 1))) == 2
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        for key, val in perm.threepats().items():
            assert perm.count_occurrences_of(key) == val
        for key, val in perm.fourpats().items():
            assert perm.count_occurrences_of(key) == val


def test_occurrences_in():
    assert list(Perm([]).occurrences_in(Perm([4, 1, 2, 3, 0]))) == [()]
    assert sorted(Perm([0]).occurrences_in(Perm([4, 1, 2, 3, 0]))) == [
        (0,),
        (1,),
        (2,),
        (3,),
        (4,),
    ]
    assert sorted(Perm([0, 1]).occurrences_in(Perm([4, 1, 2, 3, 0]))) == [
        (1, 2),
        (1, 3),
        (2, 3),
    ]
    assert sorted(Perm([1, 0]).occurrences_in(Perm([4, 1, 2, 3, 0]))) == [
        (0, 1),
        (0, 2),
        (0, 3),
        (0, 4),
        (1, 4),
        (2, 4),
        (3, 4),
    ]
    assert list(Perm([4, 1, 2, 3, 0]).occurrences_in(Perm([]))) == []
    assert list(Perm([4, 1, 2, 3, 0]).occurrences_in(Perm([1, 0]))) == []
    # Test with colours
    assert sorted(
        Perm([0]).occurrences_in(Perm([4, 1, 2, 3, 0]), [1], [0, 0, 1, 1, 0])
    ) == [(2,), (3,)]
    assert sorted(
        Perm([1, 0]).occurrences_in(Perm([4, 1, 2, 3, 0]), [1, 0], [1, 0, 1, 2, 0])
    ) == [(0, 1), (0, 4), (2, 4)]


def test_apply():
    with pytest.raises(ValueError):
        Perm((1, 2, 4, 0, 3, 5)).apply(Perm((0, 2, 1, 3)))
    with pytest.raises(ValueError):
        Perm().apply(Perm((0,)))

    for i in range(100):
        n = random.randint(0, 20)
        lst = [random.randint(0, 10000) for _ in range(n)]
        perm = Perm.random(n)
        res = list(perm.apply(lst))
        for j, k in enumerate(perm.inverse()):
            assert lst[j] == res[k]


def test_direct_sum():
    p1 = Perm((0, 1, 3, 2))
    p2 = Perm((0, 4, 2, 1, 3))
    p3 = Perm((2, 0, 1))
    p4 = Perm((0,))
    p5 = Perm()
    # All together
    result = p1 + p2 + p3 + p4 + p5
    expected = Perm((0, 1, 3, 2, 4, 8, 6, 5, 7, 11, 9, 10, 12))
    assert result == expected
    # Two
    result = p1 + p3
    expected = Perm((0, 1, 3, 2, 6, 4, 5))
    assert result == expected
    # None
    assert p1.direct_sum() == p1
    # Arguments not a permutation
    with pytest.raises(TypeError):
        p1 + None
    with pytest.raises(TypeError):
        p1.direct_sum(p2, None)
    with pytest.raises(TypeError):
        p1.direct_sum(1237)
    with pytest.raises(TypeError):
        p5 + "hahaha"


def test_skew_sum():
    p1 = Perm((0, 1, 3, 2))
    p2 = Perm((0, 4, 2, 1, 3))
    p3 = Perm((2, 0, 1))
    p4 = Perm((0,))
    p5 = Perm()
    # All together
    result = p1.skew_sum(p2, p3, p4, p5)
    expected = Perm((9, 10, 12, 11, 4, 8, 6, 5, 7, 3, 1, 2, 0))
    assert result == expected
    # Two
    result = p1 - p3
    expected = Perm((3, 4, 6, 5, 2, 0, 1))
    assert result == expected
    # None
    assert p1.skew_sum(), p1

    with pytest.raises(TypeError):
        p1 - None
    with pytest.raises(TypeError):
        p2.skew_sum(p2, None)
    with pytest.raises(TypeError):
        p3.skew_sum(1237)
    with pytest.raises(TypeError):
        p5 - "hahaha"


def test_compose():
    p0 = Perm()

    p1 = Perm((0, 3, 1, 2))
    p2 = Perm((2, 1, 0, 3))
    p3 = Perm((1, 3, 0, 2))

    p4 = Perm((1, 0, 2))
    p5 = Perm((0, 1, 2))
    p6 = Perm((2, 1, 0))
    p7 = Perm((2, 0, 1))

    assert p0.compose() == p0
    assert p1.compose() == p1
    assert p4.compose() == p4

    assert p1 * p2 == p3
    assert p1 * p3 == Perm((3, 2, 0, 1))
    assert p2 * p1 == Perm((2, 3, 1, 0))

    assert p4 * p5 * p6 == p7
    assert p5 * p6 * p7 * p4 == p7

    with pytest.raises(TypeError):
        p1.compose(None)
    with pytest.raises(TypeError):
        p1 * None
    with pytest.raises(TypeError):
        p2.compose(p2, None)
    with pytest.raises(TypeError):
        p3.compose(1237)
    with pytest.raises(TypeError):
        p5 * ("hahaha")
    with pytest.raises(TypeError):
        p5 * (p1,)

    with pytest.raises(ValueError):
        p1.compose(p0)
    with pytest.raises(ValueError):
        p0.compose(p5)
    with pytest.raises(ValueError):
        p2.compose(p3, p0)
    with pytest.raises(ValueError):
        p4.compose(p5, p6, p1)


def test_insert():
    assert Perm((0, 1)).insert() == Perm((0, 1, 2))
    assert Perm((0, 1)).insert(0) == Perm((2, 0, 1))
    assert Perm((0, 1)).insert(1) == Perm((0, 2, 1))
    assert Perm((0, 1)).insert(2) == Perm((0, 1, 2))

    assert Perm((2, 0, 1)).insert(2, 1) == Perm((3, 0, 1, 2))
    assert Perm((0, 3, 1, 2)).insert(3, 4) == Perm((0, 3, 1, 4, 2))
    assert Perm((0, 3, 1, 2)).insert(3, 3) == Perm((0, 4, 1, 3, 2))
    assert Perm((0, 3, 1, 2)).insert(1, 0) == Perm((1, 0, 4, 2, 3))

    with pytest.raises(TypeError):
        Perm((2, 1, 0, 3)).insert(new_element="hehe")
    with pytest.raises(TypeError):
        Perm((2, 1, 0, 3)).insert(3, "hehe")

    with pytest.raises(ValueError):
        Perm((2, 1, 0, 3)).insert(2, 100)
    with pytest.raises(ValueError):
        Perm((2, 1, 0, 3)).insert(0, 5)
    with pytest.raises(ValueError):
        Perm((2, 1, 0, 3)).insert(3, -3)


def test_remove():
    assert Perm().remove() == Perm()
    assert Perm((2, 0, 1)).remove() == Perm((0, 1))
    assert Perm((3, 0, 1, 2)).remove(0) == Perm((0, 1, 2))
    assert Perm((2, 0, 1)).remove(2) == Perm((1, 0))
    assert Perm((0,)).remove(0) == Perm()

    assert Perm((3, 0, 1, 2)).remove(3) == Perm((2, 0, 1))
    assert Perm((0, 3, 1, 4, 2)).remove(3) == Perm((0, 3, 1, 2))
    assert Perm((0, 4, 1, 3, 2)).remove(4) == Perm((0, 3, 1, 2))
    assert Perm((1, 0, 4, 2, 3)).remove(0) == Perm((0, 3, 1, 2))
    assert Perm((1, 0, 4, 2, 3)).remove() == Perm((1, 0, 2, 3))
    assert Perm((0, 4, 1, 3, 2)).remove() == Perm((0, 1, 3, 2))


def test_remove_element():
    assert Perm().remove() == Perm()
    assert Perm((3, 0, 1, 2)).remove_element() == Perm((0, 1, 2))
    assert Perm((3, 0, 2, 1)).remove_element(0) == Perm((2, 1, 0))
    assert Perm((3, 0, 1, 2)).remove_element(1) == Perm((2, 0, 1))
    assert Perm((0, 3, 1, 4, 2)).remove_element(4) == Perm((0, 3, 1, 2))
    assert Perm((0, 4, 1, 3, 2)).remove_element(2) == Perm((0, 3, 1, 2))
    assert Perm((1, 0, 4, 2, 3)).remove_element(1) == Perm((0, 3, 1, 2))
    assert Perm((1, 0, 4, 2, 3)).remove_element() == Perm((1, 0, 2, 3))
    assert Perm((0, 4, 1, 3, 2)).remove_element() == Perm((0, 1, 3, 2))

    with pytest.raises(ValueError):
        Perm((1, 0, 4, 2, 3)).remove_element(5)
    with pytest.raises(ValueError):
        Perm((1, 0, 4, 2, 3)).remove_element(51)
    with pytest.raises(ValueError):
        Perm((1, 0, 4, 2, 3)).remove_element(-5)

    with pytest.raises(TypeError):
        Perm((2, 1, 0, 3)).remove_element("ble")
    with pytest.raises(TypeError):
        Perm((2, 1, 0, 3)).remove_element([0])


def test_inflate():
    # TODO: make proper tests when the Perm.inflate has been implemented.
    assert Perm((0, 1)).inflate([Perm((1, 0)), Perm((2, 1, 0))]) == Perm(
        (1, 0, 4, 3, 2)
    )
    assert Perm((1, 0, 2)).inflate([None, Perm((0, 1)), Perm((0, 1))]) == Perm(
        (2, 0, 1, 3, 4)
    )
    assert Perm((0, 1)).inflate([Perm(), Perm()]) == Perm()

    with pytest.raises(TypeError):
        Perm((0, 1, 2, 3)).inflate(237)
    with pytest.raises(TypeError):
        Perm((0, 1, 2, 3)).inflate("hehe")


# TODO: The following three functions have yet to be implemented
@pytest.mark.xfail
def test_contract_inc_bonds():
    assert Perm().contract_inc_bonds() == Perm()
    assert Perm((0,)).contract_inc_bonds() == Perm((0,))
    assert Perm((1, 2)).contract_inc_bonds() == Perm((0,))
    assert Perm((1, 2, 3)).contract_inc_bonds() == Perm((0,))
    assert Perm((3, 2, 1)).contract_inc_bonds() == Perm((3, 2, 1))

    assert Perm((0, 3, 1, 4, 2)).contract_inc_bonds() == Perm((0, 3, 1, 4, 2))
    assert Perm((1, 0, 4, 2, 3)).contract_inc_bonds() == Perm((1, 0, 3, 2))
    assert Perm((1, 0, 2, 3, 4)).contract_inc_bonds() == Perm((1, 0, 2))
    assert Perm((0, 4, 1, 2, 3)).contract_inc_bonds() == Perm((0, 2, 2))


@pytest.mark.xfail
def test_contract_dec_bonds():
    assert Perm().contract_dec_bonds() == Perm()
    assert Perm((0,)).contract_dec_bonds() == Perm((0,))
    assert Perm((2, 1)).contract_dec_bonds() == Perm((0,))
    assert Perm((3, 2, 1)).contract_dec_bonds() == Perm((0,))
    assert Perm((1, 2, 3)).contract_inc_bonds() == Perm((1, 2, 3))

    assert Perm((0, 3, 1, 4, 2)).contract_dec_bonds() == Perm((0, 3, 1, 4, 2))
    assert Perm((0, 4, 3, 2, 1)).contract_dec_bonds() == Perm((0, 1))
    assert Perm((1, 0, 4, 2, 3)).contract_dec_bonds() == Perm((0, 3, 1, 2))
    assert Perm((0, 4, 1, 3, 2)).contract_dec_bonds() == Perm((0, 3, 1, 2))


@pytest.mark.xfail
def test_contract_bonds():
    assert Perm().contract_bonds() == Perm()
    assert Perm((0,)).contract_bonds() == Perm((0,))
    assert Perm((2, 1)).contract_bonds() == Perm((0,))
    assert Perm((1, 2)).contract_bonds() == Perm((0,))
    assert Perm((3, 2, 1)).contract_bonds() == Perm((0,))
    assert Perm((1, 2, 3)).contract_bonds() == Perm((0,))

    assert Perm((0, 3, 1, 4, 2)).contract_bonds() == Perm((0, 3, 1, 4, 2))
    assert Perm((0, 4, 3, 2, 1)).contract_bonds() == Perm((0, 1))
    assert Perm((1, 0, 4, 2, 3)).contract_bonds() == Perm((0, 2, 1))
    assert Perm((0, 4, 1, 3, 2)).contract_bonds() == Perm((0, 3, 1, 2))
    assert Perm((1, 0, 2, 3, 4)).contract_bonds() == Perm((0, 1, 2))


def test_inverse():
    for i in range(10):
        assert Perm(range(i)) == Perm(range(i)).inverse()
    assert Perm([2, 1, 3, 0]) == Perm([3, 1, 0, 2]).inverse()
    assert Perm((1, 2, 5, 0, 3, 4)).inverse() == Perm((3, 0, 1, 4, 5, 2))
    assert (
        Perm([4, 3, 1, 6, 5, 7, 8, 0, 2]) == Perm([7, 2, 8, 1, 0, 4, 3, 5, 6]).inverse()
    )


def test_reverse():
    assert Perm([5, 2, 3, 0, 4, 6, 1]) == Perm([1, 6, 4, 0, 3, 2, 5]).reverse()
    assert Perm([7, 1, 0, 2, 4, 6, 3, 5]) == Perm([5, 3, 6, 4, 2, 0, 1, 7]).reverse()


def test_complement():
    assert Perm().complement() == Perm()
    assert Perm((0,)).complement() == Perm((0,))
    assert Perm((1, 2, 3, 0, 4)).complement() == Perm((3, 2, 1, 4, 0))
    assert Perm((2, 0, 1)).complement() == Perm((0, 2, 1))
    assert Perm([6, 4, 2, 3, 1, 0, 5]).complement() == Perm([0, 2, 4, 3, 5, 6, 1])
    assert Perm([2, 5, 6, 3, 7, 4, 0, 1]).complement() == Perm([5, 2, 1, 4, 0, 3, 7, 6])


def test_reverse_complement():
    for _ in range(100):
        perm = Perm.random(random.randint(0, 20))
        assert perm.reverse_complement() == perm.rotate().rotate()


def test_rotate_right():
    for i in range(10):
        assert Perm(range(i - 1, -1, -1)) == Perm(range(i)).rotate_right()
    assert Perm([2, 1, 3, 4, 0, 5, 6]) == Perm([6, 5, 3, 2, 0, 1, 4]).rotate_right()
    assert (
        Perm([4, 5, 3, 1, 7, 0, 2, 6]) == Perm([4, 7, 1, 0, 2, 6, 3, 5]).rotate_right()
    )
    assert Perm([0, 1, 2]) == Perm([2, 1, 0]).rotate_right(5)
    assert Perm([4, 5, 3, 1, 0, 2]) == Perm([4, 5, 3, 1, 0, 2]).rotate_right(4)


def test_rotate_left():
    for i in range(10):
        assert Perm(list(range(i - 1, -1, -1))) == Perm(range(i)).rotate_right()
    assert Perm([6, 5, 3, 2, 0, 1, 4]).rotate_left() == Perm(
        [6, 5, 3, 2, 0, 1, 4]
    ).rotate_right(3)
    assert Perm([6, 5, 3, 2, 0, 1, 4]).rotate_left() == Perm(
        [6, 5, 3, 2, 0, 1, 4]
    ).rotate_right(-1)
    assert Perm([4, 7, 1, 0, 2, 6, 3, 5]).rotate_left() == Perm(
        [4, 7, 1, 0, 2, 6, 3, 5]
    ).rotate_right(7)
    assert Perm([]).rotate_left() == Perm([]).rotate_left(123)


def test_shift_left():
    assert Perm([]) == Perm([]).shift_left()
    assert Perm([0]) == Perm([0]).shift_left()
    assert Perm([0, 1, 2, 3, 4, 5, 6, 7]) == Perm([0, 1, 2, 3, 4, 5, 6, 7]).shift_left(
        0
    )
    assert Perm([0, 1, 2, 3, 4, 5, 6, 7]) == Perm([5, 6, 7, 0, 1, 2, 3, 4]).shift_left(
        3
    )
    assert Perm([0, 1, 2, 3, 4, 5, 6, 7]) == Perm([0, 1, 2, 3, 4, 5, 6, 7]).shift_left(
        800
    )
    assert Perm([0, 1, 2, 3, 4, 5, 6, 7]) == Perm([5, 6, 7, 0, 1, 2, 3, 4]).shift_left(
        403
    )
    assert Perm([0, 1, 2, 3, 4, 5, 6, 7]) == Perm([0, 1, 2, 3, 4, 5, 6, 7]).shift_left(
        -8
    )
    assert Perm([0, 1, 2, 3, 4, 5, 6, 7]) == Perm([5, 6, 7, 0, 1, 2, 3, 4]).shift_left(
        -5
    )


def test_shift_down():
    assert Perm([]) == Perm([]).shift_down(1000)
    assert Perm([0]) == Perm([0]).shift_down(10)
    assert Perm([1, 0, 4, 2, 3, 5]) == Perm([2, 1, 5, 3, 4, 0]).shift_down()
    assert Perm([1, 0, 4, 2, 3, 5]) == Perm([2, 1, 5, 3, 4, 0]).shift_down(13)
    assert Perm([1, 0, 4, 2, 3, 5]) == Perm([5, 4, 2, 0, 1, 3]).shift_down(-2)
    assert Perm([1, 0, 4, 2, 3, 5]) == Perm([5, 4, 2, 0, 1, 3]).shift_down(-8)


def test_flip_horizontal():
    assert Perm().flip_horizontal() == Perm()
    assert Perm((0,)).flip_horizontal() == Perm((0,))
    assert Perm((1, 2, 3, 0, 4)).flip_horizontal() == Perm((3, 2, 1, 4, 0))
    assert Perm((2, 0, 1)).flip_horizontal() == Perm((0, 2, 1))
    assert Perm([6, 4, 2, 3, 1, 0, 5]).flip_horizontal() == Perm([0, 2, 4, 3, 5, 6, 1])
    assert Perm([2, 5, 6, 3, 7, 4, 0, 1]).flip_horizontal() == Perm(
        [5, 2, 1, 4, 0, 3, 7, 6]
    )


def test_flip_vertical():
    assert Perm().flip_vertical() == Perm()
    assert Perm((0,)).flip_vertical() == Perm((0,))
    assert Perm((0, 1)).flip_vertical() == Perm((1, 0))
    assert Perm((1, 2, 5, 0, 3, 4)).flip_vertical() == Perm((4, 3, 0, 5, 2, 1))
    assert Perm([5, 2, 3, 0, 4, 6, 1]).reverse() == Perm([1, 6, 4, 0, 3, 2, 5])
    assert Perm([7, 1, 0, 2, 4, 6, 3, 5]).reverse() == Perm([5, 3, 6, 4, 2, 0, 1, 7])


def test_flip_diagonal():
    for i in range(10):
        assert Perm(range(i)) == Perm(range(i)).flip_diagonal()
    assert Perm([2, 1, 3, 0]) == Perm([3, 1, 0, 2]).flip_diagonal()
    assert Perm((1, 2, 5, 0, 3, 4)).flip_diagonal() == Perm((3, 0, 1, 4, 5, 2))
    assert (
        Perm([4, 3, 1, 6, 5, 7, 8, 0, 2])
        == Perm([7, 2, 8, 1, 0, 4, 3, 5, 6]).flip_diagonal()
    )


def test_flip_antidiagonal():
    for i in range(100):
        perm = Perm.random(random.randint(0, 20))
        assert perm.reverse().complement().inverse() == perm.flip_antidiagonal()


def test__rotate_180():
    for _ in range(100):
        perm = Perm.random(random.randint(0, 20))
        assert perm._rotate_180() == perm.rotate().rotate()


def test_all_syms():
    assert list(sorted(Perm((0,)).all_syms())) == [Perm((0,))]
    assert list(sorted(Perm((0, 2, 1)).all_syms())) == [
        Perm((0, 2, 1)),
        Perm((1, 0, 2)),
        Perm((1, 2, 0)),
        Perm((2, 0, 1)),
    ]


@pytest.mark.xfail
def test_is_representative():
    # TODO: write proper tests when the function is working
    assert Perm().is_representative()
    assert Perm((0,)).is_representative()


def test_fixed_points():
    assert list(Perm().fixed_points()) == []
    assert list(Perm((0, 2, 1)).fixed_points()) == [0]
    assert list(Perm((5, 4, 3, 2, 1, 0)).fixed_points()) == []
    assert list(Perm((4, 1, 0, 3, 2, 5)).fixed_points()) == [1, 3, 5]
    assert list(Perm((0, 1, 2, 3, 4, 5)).fixed_points()) == [0, 1, 2, 3, 4, 5]


def test_count_fixed_points():
    assert Perm().count_fixed_points() == 0
    assert Perm((0, 2, 1)).count_fixed_points() == 1
    assert Perm((5, 4, 3, 2, 1, 0)).count_fixed_points() == 0
    assert Perm((4, 1, 0, 3, 2, 5)).count_fixed_points() == 3
    assert Perm((0, 1, 2, 3, 4, 5)).count_fixed_points() == 6


def test_is_skew_decomposable():
    p1 = Perm((0, 1, 3, 2))
    p2 = Perm((0, 4, 2, 1, 3))
    p3 = Perm((2, 0, 1))
    p4 = Perm((0,))
    p5 = Perm()

    assert p1.skew_sum(p2).is_skew_decomposable()
    assert p1.skew_sum(p2, p3).is_skew_decomposable()
    assert p1.skew_sum(p2, p3, p4).is_skew_decomposable()
    assert p2.skew_sum(p3, p4).is_skew_decomposable()
    assert not p5.is_skew_decomposable()

    assert not Perm((0, 1, 2, 3, 4)).is_skew_decomposable()
    assert not Perm((0, 5, 4, 3, 2, 1)).is_skew_decomposable()


def test_skew_decomposition():
    p1 = Perm((0, 1, 3, 2))
    p2 = Perm((0, 4, 2, 1, 3))
    p3 = Perm((2, 0, 1))
    p4 = Perm((0,))
    p5 = Perm()
    p6 = Perm((0, 1))

    assert p3.skew_decomposition() == [p4, p6]
    assert p1.skew_sum(p2).skew_decomposition() == [p1, p2]
    assert p1.skew_sum(p2, p3).skew_decomposition() == [p1, p2, p4, p6]
    assert p1.skew_sum(p2, p3, p4).skew_decomposition() == [p1, p2, p4, p6, p4]
    assert p5.skew_decomposition() == []
    assert p4.skew_decomposition() == [p4]
    assert p1.skew_decomposition() == [p1]
    assert Perm((0, 1, 2, 3, 4)).skew_decomposition() == [Perm((0, 1, 2, 3, 4))]
    assert Perm((0, 5, 4, 3, 2, 1)).skew_decomposition() == [Perm((0, 5, 4, 3, 2, 1))]


def test_is_sum_decomposable():
    p1 = Perm((0, 1, 3, 2))
    p2 = Perm((4, 2, 1, 3, 0))
    p3 = Perm((2, 0, 1))
    p4 = Perm((0,))
    p5 = Perm()

    assert p1.is_sum_decomposable()
    assert p1.direct_sum(p2, p3).is_sum_decomposable()
    assert p1.direct_sum(p2, p3, p4).is_sum_decomposable()
    assert p1.direct_sum(p2, p3, p4, p5).is_sum_decomposable()

    assert not p2.is_sum_decomposable()
    assert not p3.is_sum_decomposable()
    assert not Perm((4, 3, 2, 1, 0)).is_sum_decomposable()


def test_sum_decomposition():
    p1 = Perm((0, 1, 3, 2))
    p2 = Perm((4, 2, 1, 3, 0))
    p3 = Perm((2, 0, 1))
    p4 = Perm((0,))
    p5 = Perm()
    p6 = Perm((1, 0))

    assert p1.sum_decomposition() == [p4, p4, p6]
    assert p1.direct_sum(p2, p3).sum_decomposition() == [p4, p4, p6, p2, p3]
    assert p1.direct_sum(p2, p3, p4).sum_decomposition() == [
        p4,
        p4,
        p6,
        p2,
        p3,
        p4,
    ]
    assert p4.sum_decomposition() == [p4]
    assert p5.sum_decomposition() == []
    assert p2.sum_decomposition() == [p2]
    assert p3.sum_decomposition() == [p3]
    assert Perm((4, 3, 2, 1, 0)).sum_decomposition() == [Perm((4, 3, 2, 1, 0))]


def test_descent_set():
    assert Perm().descent_set() == []
    assert Perm((0, 1, 2, 3)).descent_set() == []
    assert Perm((3, 2, 1, 0)).descent_set() == [0, 1, 2]
    assert Perm((2, 1, 0, 4, 3, 5)).descent_set() == [0, 1, 3]
    assert Perm((1, 2, 3, 0, 6, 5, 4)).descent_set() == [2, 4, 5]
    assert Perm((3, 1, 4, 5, 0, 7, 6, 2)).descent_set() == [0, 3, 5, 6]


def test_count_descents():
    assert Perm().count_descents() == 0
    assert Perm((0, 1, 2, 3)).count_descents() == 0
    assert Perm((3, 2, 1, 0)).count_descents() == 3
    assert Perm((2, 1, 0, 4, 3, 5)).count_descents() == 3
    assert Perm((1, 2, 3, 0, 6, 5, 4)).count_descents() == 3
    assert Perm((3, 1, 4, 5, 0, 7, 6, 2)).count_descents() == 4


def test_ascents():
    assert list(Perm().ascents()) == []
    assert list(Perm((0, 1, 2, 3)).ascents()) == [0, 1, 2]
    assert list(Perm((3, 2, 1, 0)).ascents()) == []
    assert list(Perm((2, 1, 0, 4, 3, 5)).ascents()) == [2, 4]
    assert list(Perm((1, 2, 3, 0, 6, 5, 4)).ascents()) == [0, 1, 3]
    assert list(Perm((3, 1, 4, 5, 0, 7, 6, 2)).ascents()) == [1, 2, 4]


def test_ascent_set():
    assert Perm().ascent_set() == []
    assert Perm((0, 1, 2, 3)).ascent_set() == [0, 1, 2]
    assert Perm((3, 2, 1, 0)).ascent_set() == []
    assert Perm((2, 1, 0, 4, 3, 5)).ascent_set() == [2, 4]
    assert Perm((1, 2, 3, 0, 6, 5, 4)).ascent_set() == [0, 1, 3]
    assert Perm((3, 1, 4, 5, 0, 7, 6, 2)).ascent_set() == [1, 2, 4]


def test_count_ascents():
    assert Perm().count_ascents() == 0
    assert Perm((0, 1, 2, 3)).count_ascents() == 3
    assert Perm((3, 2, 1, 0)).count_ascents() == 0
    assert Perm((2, 1, 0, 4, 3, 5)).count_ascents() == 2
    assert Perm((1, 2, 3, 0, 6, 5, 4)).count_ascents() == 3
    assert Perm((3, 1, 4, 5, 0, 7, 6, 2)).count_ascents() == 3


def test_peaks():
    assert list(Perm().peaks()) == []
    assert list(Perm((0, 1, 2, 3)).peaks()) == []
    assert list(Perm((0, 1, 2, 4, 3)).peaks()) == [3]
    assert list(Perm((2, 1, 0, 4, 3, 5)).peaks()) == [3]
    assert list(Perm((1, 2, 3, 0, 6, 5, 4)).peaks()) == [2, 4]


def test_peak_list():
    assert Perm().peak_list() == []
    assert Perm((0, 1, 2, 3)).peak_list() == []
    assert Perm((0, 1, 2, 4, 3)).peak_list() == [3]
    assert Perm((2, 1, 0, 4, 3, 5)).peak_list() == [3]
    assert Perm((1, 2, 3, 0, 6, 5, 4)).peak_list() == [2, 4]


def test_count_peaks():
    assert Perm().count_peaks() == 0
    assert Perm((0, 1, 2, 3)).count_peaks() == 0
    assert Perm((2, 1, 0, 4, 3, 5)).count_peaks() == 1
    assert Perm((1, 2, 3, 0, 6, 5, 4)).count_peaks() == 2


def test_valleys():
    assert list(Perm().valleys()) == []
    assert list(Perm((0, 1, 2, 3)).valleys()) == []
    assert list(Perm((2, 1, 0, 4, 3, 5)).valleys()) == [2, 4]
    assert list(Perm((1, 2, 3, 0, 6, 5, 4)).valleys()) == [3]
    assert list(Perm((2, 1, 3, 0, 6, 4, 5)).valleys()) == [1, 3, 5]


def test_valley_list():
    assert Perm().valley_list() == []
    assert Perm((0, 1, 2, 3)).valley_list() == []
    assert Perm((2, 1, 0, 4, 3, 5)).valley_list() == [2, 4]
    assert Perm((1, 2, 3, 0, 6, 5, 4)).valley_list() == [3]
    assert Perm((2, 1, 3, 0, 6, 4, 5)).valley_list() == [1, 3, 5]


def test_count_valleys():
    assert Perm().count_valleys() == 0
    assert Perm((0, 1, 2, 3)).count_valleys() == 0
    assert Perm((2, 1, 0, 4, 3, 5)).count_valleys() == 2
    assert Perm((1, 2, 3, 0, 6, 5, 4)).count_valleys() == 1
    assert Perm((2, 1, 3, 0, 6, 4, 6)).count_valleys() == 3


def test_bends():
    assert list(Perm().bends()) == []
    assert list(Perm((0, 1)).bends()) == []
    assert list(Perm((2, 0, 1)).bends()) == [1]
    assert list(Perm((5, 3, 4, 0, 2, 1)).bends()) == [
        1,
        2,
        3,
        4,
    ]
    assert list(Perm((4, 3, 5, 7, 6, 9, 1, 2, 8, 0)).bends()) == [1, 3, 4, 5, 6, 8]
    assert list(Perm((6, 4, 3, 0, 1, 7, 2, 5, 8, 9)).bends()) == [3, 5, 6]


def test_bend_list():
    assert Perm().bend_list() == []
    assert Perm((0, 1)).bend_list() == []
    assert Perm((2, 0, 1)).bend_list() == [1]
    assert Perm((5, 3, 4, 0, 2, 1)).bend_list() == [
        1,
        2,
        3,
        4,
    ]
    assert Perm((4, 3, 5, 7, 6, 9, 1, 2, 8, 0)).bend_list() == [1, 3, 4, 5, 6, 8]
    assert Perm((6, 4, 3, 0, 1, 7, 2, 5, 8, 9)).bend_list() == [3, 5, 6]


def test_order():
    assert Perm().order() == 1
    assert Perm((0,)).order() == 1
    perm = Perm((4, 5, 2, 0, 6, 1, 3))
    args = tuple(perm for _ in range(perm.order() - 1))
    assert perm.compose(*args).is_identity()
    for _ in range(100):
        perm = Perm.random(random.randint(2, 20))
        args = tuple(perm for _ in range(perm.order() - 1))
        assert perm.compose(*args).is_identity()


def test_ltrmin():
    assert Perm().ltrmin() == []
    assert Perm((0,)).ltrmin() == [0]
    assert Perm((2, 4, 3, 0, 1)).ltrmin() == [0, 3]
    for _ in range(100):
        perm = Perm.random(random.randint(2, 20))
        ltrmin_list = perm.ltrmin()
        assert ltrmin_list[0] == 0
        for i in range(len(ltrmin_list)):
            for j in range(ltrmin_list[i] + 1, len(perm)):
                if perm[j] < perm[ltrmin_list[i]]:
                    assert ltrmin_list[i + 1] == j
                    break


def test_rtlmin():
    assert Perm().rtlmin() == []
    assert Perm((0,)).rtlmin() == [0]
    assert Perm((2, 4, 3, 0, 1)).rtlmin() == [3, 4]
    for _ in range(100):
        perm = Perm.random(random.randint(2, 20))
        rtlmin_list = perm.rtlmin()
        rtlmin_list.reverse()
        assert rtlmin_list[0] == len(perm) - 1
        for i in range(len(rtlmin_list)):
            for j in range(rtlmin_list[i] - 1, -1, -1):
                if perm[j] < perm[rtlmin_list[i]]:
                    assert rtlmin_list[i + 1] == j
                    break


def test_ltrmax():
    assert Perm().ltrmax() == []
    assert Perm((0,)).ltrmax() == [0]
    assert Perm((2, 0, 4, 1, 5, 3)).ltrmax() == [0, 2, 4]
    for _ in range(100):
        perm = Perm.random(random.randint(2, 20))
        ltrmax_list = perm.ltrmax()
        assert ltrmax_list[0] == 0
        for i in range(len(ltrmax_list)):
            for j in range(ltrmax_list[i] + 1, len(perm)):
                if perm[j] > perm[ltrmax_list[i]]:
                    assert ltrmax_list[i + 1] == j
                    break


def test_rtlmax():
    assert Perm().rtlmax() == []
    assert Perm((0,)).rtlmax() == [0]
    assert Perm((2, 4, 3, 0, 1)).rtlmax() == [1, 2, 4]
    for _ in range(100):
        perm = Perm.random(random.randint(2, 20))
        rtlmax_list = perm.rtlmax()
        rtlmax_list.reverse()
        assert rtlmax_list[0] == len(perm) - 1
        for i in range(len(rtlmax_list)):
            for j in range(rtlmax_list[i] - 1, -1, -1):
                if perm[j] > perm[rtlmax_list[i]]:
                    assert rtlmax_list[i + 1] == j
                    break


def test_count_ltrmin():
    assert Perm().count_ltrmin() == 0
    assert Perm((0,)).count_ltrmin() == 1
    assert Perm((0, 1)).count_ltrmin() == 1
    assert Perm((4, 0, 6, 3, 2, 1, 5, 8, 7, 9)).count_ltrmin() == 2
    assert Perm((5, 7, 3, 4, 6, 8, 2, 9, 0, 1)).count_ltrmin() == 4
    assert Perm((3, 1, 5, 9, 6, 4, 7, 8, 2, 0)).count_ltrmin() == 3


def test_inversions():
    assert list(Perm().inversions()) == []
    assert list(Perm((0,)).inversions()) == []
    assert list(Perm((0, 1)).inversions()) == []
    assert list(Perm((1, 0)).inversions()) == [(0, 1)]
    assert list(Perm((3, 0, 2, 1)).inversions()) == [(0, 1), (0, 2), (0, 3), (2, 3)]


def test_count_inversions():
    assert Perm().count_inversions() == 0
    assert Perm((0,)).count_inversions() == 0
    assert Perm((0, 1)).count_inversions() == 0
    assert Perm((1, 0)).count_inversions() == 1
    for _ in range(50):
        perm = Perm.random(random.randint(0, 20))
        invs = 0
        for i in range(len(perm)):
            for j in range(i + 1, len(perm)):
                if perm[i] > perm[j]:
                    invs += 1
        assert perm.count_inversions() == invs


def test_non_inversions():
    assert list(Perm().non_inversions()) == []
    assert list(Perm((0,)).non_inversions()) == []
    assert list(Perm((0, 1)).non_inversions()) == [(0, 1)]
    assert list(Perm((1, 0)).non_inversions()) == []
    assert list(Perm((3, 0, 2, 1, 4)).non_inversions()) == [
        (0, 4),
        (1, 2),
        (1, 3),
        (1, 4),
        (2, 4),
        (3, 4),
    ]


def test_count_non_inversions():
    assert Perm().count_non_inversions() == 0
    assert Perm((0,)).count_non_inversions() == 0
    assert Perm((0, 1)).count_non_inversions() == 1
    assert Perm((1, 0)).count_non_inversions() == 0
    for _ in range(50):
        perm = Perm.random(random.randint(0, 20))
        invs = 0
        for i in range(len(perm)):
            for j in range(i + 1, len(perm)):
                if perm[i] < perm[j]:
                    invs += 1
        assert perm.count_non_inversions() == invs


def test_count_bonds():
    assert Perm().count_bonds() == 0
    assert Perm((0,)).count_bonds() == 0
    for _ in range(50):
        perm = Perm.random(random.randint(0, 20))
        bons = 0
        for i in range(len(perm) - 1):
            if perm[i] + 1 == perm[i + 1] or perm[i] == perm[i + 1] + 1:
                bons += 1
        assert bons == perm.count_bonds()


def test_major_index():
    assert Perm().major_index() == 0
    assert Perm((0,)).major_index() == 0
    assert Perm((0, 2, 1)).major_index() == 2
    assert Perm((3, 1, 2, 4, 0)).major_index() == 5
    assert Perm((3, 0, 7, 4, 1, 2, 5, 6)).major_index() == 8
    assert Perm((7, 0, 4, 3, 1, 5, 2, 6)).major_index() == 14


def test_minimum_gapsize():
    assert Perm((0, 1)).min_gapsize() == 2
    assert Perm((2, 0, 3, 1)).min_gapsize() == 3
    assert Perm((7, 2, 5, 3, 4, 1, 6, 0)).min_gapsize() == 2
    assert Perm((5, 2, 0, 3, 6, 4, 7, 1)).min_gapsize() == 3
    assert Perm((7, 2, 0, 4, 9, 5, 8, 1, 6, 3)).min_gapsize() == 3


def test_longestruns_descending():
    assert Perm().longestruns_descending() == (0, [])
    assert Perm((3, 2, 1, 0)).longestruns_descending() == (4, [0])
    assert Perm((0, 1, 2, 3, 4)).longestruns_descending() == (1, [0, 1, 2, 3, 4])
    assert Perm((3, 9, 7, 0, 2, 5, 8, 6, 4, 1)).longestruns_descending() == (4, [6])
    assert Perm((0, 3, 1, 4, 7, 2, 8, 5, 6, 9)).longestruns_descending() == (
        2,
        [1, 4, 6],
    )


def test_longestruns():
    assert Perm().longestruns() == (0, [])
    assert Perm((0, 1, 2, 3)).longestruns() == (4, [0])
    assert Perm((4, 3, 2, 1, 0)).longestruns() == (1, [0, 1, 2, 3, 4])
    assert Perm((8, 1, 7, 5, 6, 2, 9, 3, 0, 4)).longestruns() == (2, [1, 3, 5, 8])
    assert Perm((1, 2, 3, 4, 6, 0, 9, 7, 8, 5)).longestruns() == (5, [0])


def test_length_of_longestrun_ascending():
    assert Perm().length_of_longestrun_ascending() == 0
    assert Perm((2, 1, 0)).length_of_longestrun_ascending() == 1
    assert Perm((0, 1, 2)).length_of_longestrun_ascending() == 3
    assert Perm((0, 1, 2, 3, 4, 5, 6, 7, 8, 9)).length_of_longestrun_ascending() == 10
    assert Perm((6, 0, 9, 1, 4, 7, 3, 8, 5, 2)).length_of_longestrun_ascending() == 3


def test_length_of_longestrun_descending():
    assert Perm().length_of_longestrun_descending() == 0
    assert Perm((2, 1, 0)).length_of_longestrun_descending() == 3
    assert Perm((0, 1, 2)).length_of_longestrun_descending() == 1
    assert Perm((3, 5, 2, 7, 1, 8, 0, 6, 9, 4)).length_of_longestrun_descending() == 2
    assert Perm((5, 4, 8, 9, 7, 3, 2, 1, 0, 6)).length_of_longestrun_descending() == 6


def test_length_of_longestrun():
    assert Perm((6, 1, 3, 7, 4, 5, 9, 8, 0, 2)).length_of_longestrun() == 3
    assert Perm((4, 0, 9, 5, 3, 7, 1, 6, 8, 2)).length_of_longestrun() == 3
    assert Perm((1, 9, 4, 6, 0, 8, 2, 7, 5, 3)).length_of_longestrun() == 2
    assert Perm((2, 5, 8, 6, 0, 1, 3, 7, 9, 4)).length_of_longestrun() == 5


def test_count_cycles():
    for i in range(10):
        assert Perm.identity(i).count_cycles() == i
    assert Perm((2, 0, 1)).count_cycles() == 1
    assert Perm((4, 2, 7, 0, 3, 1, 6, 5)).count_cycles() == 3
    assert Perm((5, 3, 8, 1, 0, 4, 2, 7, 6)).count_cycles() == 4


def test_is_involution():
    assert Perm().is_involution()
    assert Perm((0,)).is_involution()
    for _ in range(30):
        perm = Perm.random(random.randint(0, 20))
        cyclelist = perm.cycle_decomp()
        assert perm.is_involution() == all(map(lambda x: len(x) <= 2, cyclelist))


def test_rank():
    assert list(map(lambda x: x.rank(), PermSet(5))) == list(range(34, 154))


def test_threepats():
    assert all(v == 0 for v in Perm().threepats().values())
    assert all(v == 0 for v in Perm((0,)).threepats().values())
    assert all(v == 0 for v in Perm((0, 1)).threepats().values())
    assert Perm((2, 1, 0, 3)).threepats() == {
        Perm((0, 1, 2)): 0,
        Perm((0, 2, 1)): 0,
        Perm((1, 0, 2)): 3,
        Perm((1, 2, 0)): 0,
        Perm((2, 0, 1)): 0,
        Perm((2, 1, 0)): 1,
    }
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        threepatdict = perm.threepats()
        for key, val in threepatdict.items():
            assert key.count_occurrences_in(perm) == val


def test_fourpats():
    assert all(v == 0 for v in Perm().fourpats().values())
    assert all(v == 0 for v in Perm((0,)).fourpats().values())
    assert all(v == 0 for v in Perm((0, 1)).fourpats().values())
    assert Perm((1, 0, 3, 5, 2, 4)).fourpats() == {
        Perm((0, 1, 2, 3)): 0,
        Perm((0, 1, 3, 2)): 2,
        Perm((0, 2, 1, 3)): 2,
        Perm((0, 2, 3, 1)): 2,
        Perm((0, 3, 1, 2)): 2,
        Perm((0, 3, 2, 1)): 0,
        Perm((1, 0, 2, 3)): 3,
        Perm((1, 0, 3, 2)): 3,
        Perm((1, 2, 0, 3)): 0,
        Perm((1, 2, 3, 0)): 0,
        Perm((1, 3, 0, 2)): 1,
        Perm((1, 3, 2, 0)): 0,
        Perm((2, 0, 1, 3)): 0,
        Perm((2, 0, 3, 1)): 0,
        Perm((2, 1, 0, 3)): 0,
        Perm((2, 1, 3, 0)): 0,
        Perm((2, 3, 0, 1)): 0,
        Perm((2, 3, 1, 0)): 0,
        Perm((3, 0, 1, 2)): 0,
        Perm((3, 0, 2, 1)): 0,
        Perm((3, 1, 0, 2)): 0,
        Perm((3, 1, 2, 0)): 0,
        Perm((3, 2, 0, 1)): 0,
        Perm((3, 2, 1, 0)): 0,
    }
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        fourpatdict = perm.fourpats()
        for key, val in fourpatdict.items():
            assert key.count_occurrences_in(perm) == val


def test_rank_encoding():
    assert Perm().rank_encoding() == []
    assert Perm((0,)).rank_encoding() == [0]
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        for index, val in enumerate(perm.rank_encoding()):
            invs = 0
            for i in range(index + 1, len(perm)):
                if perm[i] < perm[index]:
                    invs += 1
            assert invs == val


def test_block_decomposition():
    assert Perm().block_decomposition() == []
    assert Perm((0,)).block_decomposition() == [[]]
    assert Perm((5, 3, 0, 1, 2, 4, 7, 6)).block_decomposition() == [
        [],
        [],
        [2, 3, 6],
        [2],
        [1],
        [1],
        [0],
        [],
    ]
    assert set(Perm((4, 1, 0, 5, 2, 3)).block_decomposition(True)) == set(
        [Perm((0, 1)), Perm((1, 0))]
    )
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        blocks = perm.block_decomposition()
        patts = set(perm.block_decomposition(True))
        for length in range(len(blocks)):
            for start in blocks[length]:
                assert (
                    max(perm[start : start + length])
                    - min(perm[start : start + length])
                    == length - 1
                )
                assert Perm.to_standard(perm[start : start + length]) in patts


def test_monotone_block_decomposition():
    assert Perm().monotone_block_decomposition(True) == []
    assert Perm((0,)).monotone_block_decomposition() == []
    assert Perm((0,)).monotone_block_decomposition(True) == [(0, 0)]
    assert Perm((6, 7, 5, 3, 0, 1, 2, 4)).monotone_block_decomposition() == [
        (0, 1),
        (4, 6),
    ]
    assert Perm((0, 2, 1, 5, 6, 7, 4, 3)).monotone_block_decomposition() == [
        (1, 2),
        (3, 5),
        (6, 7),
    ]
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        monblocks = perm.monotone_block_decomposition(True)
        last = -1
        for block in monblocks:
            assert block[0] == last + 1
            last = block[1]
            assert all(
                perm[i] - perm[i - 1] == perm[block[0] + 1] - perm[block[0]]
                for i in range(block[0] + 2, block[1])
            )


def test_monotone_quotient():
    assert Perm().monotone_quotient() == Perm()
    assert Perm((0,)).monotone_quotient() == Perm((0,))
    assert Perm((0, 2, 1, 5, 6, 7, 4, 3)).monotone_quotient() == Perm((0, 1, 3, 2))
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        monblocks = tuple(
            start for (start, end) in perm.monotone_block_decomposition(True)
        )
        assert monblocks in list(perm.occurrences_of(perm.monotone_quotient()))


def test_simple_location():
    assert Perm().simple_location() == (0, 0)
    assert Perm((0,)).simple_location() == (0, 0)
    assert Perm((0, 2, 1, 5, 6, 7, 4, 3)).simple_location() == (7, 1)
    assert Perm((3, 4, 0, 7, 2, 6, 1, 5)).simple_location() == (2, 0)
    for _ in range(20):
        perm = Perm.random(random.randint(0, 20))
        length, start = perm.simple_location()
        if length != 0:
            assert (
                max(perm[start : start + length]) - min(perm[start : start + length])
                == length - 1
            )
        else:
            length = 2
        for bigger in range(length + 1, len(perm) - 1):
            for start in range(len(perm) - bigger - 1):
                assert (
                    max(perm[start : start + bigger])
                    - min(perm[start : start + bigger])
                    != bigger - 1
                )


def test_is_simple():
    assert Perm().is_simple()
    assert Perm((0,)).is_simple()
    assert not Perm((0, 1, 2)).is_simple()
    assert not Perm((0, 2, 1)).is_simple()
    assert not Perm((1, 0, 2)).is_simple()
    assert not Perm((1, 2, 0)).is_simple()
    assert not Perm((2, 0, 1)).is_simple()
    assert not Perm((2, 1, 0)).is_simple()
    assert Perm((2, 0, 3, 1)).is_simple()
    assert not Perm((3, 2, 0, 1)).is_simple()
    assert Perm((1, 3, 0, 2)).is_simple()
    assert Perm((3, 7, 2, 6, 1, 5, 0, 4)).is_simple()


def test_is_strongly_simple():
    assert Perm().is_strongly_simple()
    assert Perm((0,)).is_strongly_simple()
    assert not Perm((0, 1, 2)).is_strongly_simple()
    assert not Perm((0, 2, 1)).is_strongly_simple()
    assert not Perm((1, 0, 2)).is_strongly_simple()
    assert Perm((4, 1, 6, 3, 0, 7, 2, 5)).is_strongly_simple()


def test_coveredby():
    assert Perm().coveredby() == [Perm((0,))]
    assert sorted(Perm((0,)).coveredby()) == sorted([Perm((0, 1)), Perm((1, 0))])
    assert sorted(Perm((0, 1)).coveredby()) == sorted(
        [
            Perm((0, 2, 1)),
            Perm((1, 2, 0)),
            Perm((0, 1, 2)),
            Perm((2, 0, 1)),
            Perm((1, 0, 2)),
        ]
    )
    for _ in range(10):
        perm = Perm.random(random.randint(0, 12))
        for p in perm.coveredby():
            assert perm in p.children()


def test_call_1():
    p = Perm((0, 1, 2, 3))
    for i in range(len(p)):
        assert p(i) == i
    with pytest.raises(ValueError):
        p(-1)
    with pytest.raises(ValueError):
        p(4)
    with pytest.raises(TypeError):
        p("abc")


def test_call_2():
    p = Perm((3, 4, 0, 2, 1))
    assert p(0) == 3
    assert p(1) == 4
    assert p(2) == 0
    assert p(3) == 2
    assert p(4) == 1
    with pytest.raises(ValueError):
        p(-1)
    with pytest.raises(ValueError):
        p(5)
    with pytest.raises(TypeError):
        p([1, 2, 3])


def test_eq():
    assert Perm([]) == Perm([])
    assert Perm([0]) == Perm([0])
    assert Perm([]) != Perm([0])
    assert Perm([0]) != Perm([])
    assert not (Perm([]) != Perm([]))
    assert not (Perm([0]) != Perm([0]))
    assert not (Perm([]) == Perm([0]))
    assert not (Perm([0]) == Perm([]))
    for _ in range(100):
        a = Perm.random(random.randint(0, 10))
        b = Perm(a)
        c = Perm.random(random.randint(0, 10))
        if a == c:
            continue
        assert a == b
        assert a != c
        assert b == a
        assert c != a


def test_avoids():
    assert Perm([4, 0, 1, 2, 3]).avoids()
    assert not (Perm([4, 0, 1, 2, 3]).avoids(Perm([0, 1, 2])))
    assert not (Perm([4, 0, 1, 2, 3]).avoids(Perm([1, 0])))
    assert Perm([4, 0, 1, 2, 3]).avoids(Perm([2, 1, 0]))
    assert not (Perm([4, 0, 1, 2, 3]).avoids_set([Perm([2, 1, 0]), Perm([1, 0])]))
    assert Perm([4, 0, 1, 2, 3]).avoids_set([Perm([2, 1, 0]), Perm([1, 2, 0])])


def test_avoids_2():
    bound = 6

    def do_test(patts, expected):
        for i in range(min(len(expected), bound)):
            length = i + 1
            cnt = 0
            for p in PermSet(length):
                ok = True
                for patt in patts:
                    if not p.avoids(Perm(patt)):
                        ok = False
                        break
                if ok:
                    cnt += 1
            assert expected[i] == cnt

    do_test([[0, 1, 2]], [1, 2, 5, 14, 42, 132, 429, 1430])
    do_test([[1, 2, 0]], [1, 2, 5, 14, 42, 132, 429, 1430])
    do_test([[0, 2, 3, 1]], [1, 2, 6, 23, 103, 512, 2740, 15485])
    do_test([[1, 3, 0, 2]], [1, 2, 6, 23, 103, 512, 2740, 15485])
    do_test([[0, 1, 2, 3]], [1, 2, 6, 23, 103, 513, 2761, 15767])
    do_test([[0, 3, 2, 1]], [1, 2, 6, 23, 103, 513, 2761, 15767])
    do_test([[1, 0, 3, 2]], [1, 2, 6, 23, 103, 513, 2761, 15767])
    do_test([[0, 2, 1, 3]], [1, 2, 6, 23, 103, 513, 2762, 15793])


def test_incr_decr():
    for i in range(100):
        assert Perm(range(i)).is_increasing()
        assert Perm(range(i - 1, -1, -1)).is_decreasing()

    assert not Perm([0, 2, 1]).is_increasing()
    assert not Perm([0, 2, 1]).is_decreasing()
    assert not Perm([1, 0, 2]).is_increasing()
    assert not Perm([1, 0, 2]).is_decreasing()


def test_lt():
    # TODO: No length testing is done here
    for _ in range(30):
        l1 = list(range(10))
        l2 = list(range(10))
        random.shuffle(l1)
        random.shuffle(l2)
        if l1 < l2:
            assert Perm(l1) < Perm(l2)
        else:
            assert not (Perm(l1) < Perm(l2))
        assert not (Perm(l1) < Perm(l1))
        assert not (Perm(l2) < Perm(l2))


def test_gt():
    # TODO: No length testing is done here
    for _ in range(30):
        l1 = list(range(10))
        l2 = list(range(10))
        random.shuffle(l1)
        random.shuffle(l2)
        if l1 > l2:
            assert Perm(l1) > Perm(l2)
        else:
            assert not (Perm(l1) > Perm(l2))
        assert not (Perm(l1) > Perm(l1))
        assert not (Perm(l2) > Perm(l2))


def test_ge():
    # TODO: No length testing is done here
    for _ in range(30):
        l1 = list(range(10))
        l2 = list(range(10))
        random.shuffle(l1)
        random.shuffle(l2)
        if l1 >= l2:
            assert Perm(l1) >= Perm(l2)
        else:
            assert not (Perm(l1) >= Perm(l2))
        assert Perm(l1) >= Perm(l1)
        assert Perm(l2) >= Perm(l2)


def test_bool():
    assert Perm([0, 1, 2, 3])
    assert Perm([0])
    assert not (Perm([]))
    assert not (Perm())


def test_ascii_plot():
    assert Perm().ascii_plot() == ""
    assert Perm((0,)).ascii_plot() == " |\n-\u25cf-\n |"
    assert (
        Perm((0, 1)).ascii_plot(cell_size=2) == "  |  |\n"
        "  |  |\n"
        "--+--●--\n"
        "  |  |\n"
        "  |  |\n"
        "--●--+--\n"
        "  |  |\n"
        "  |  |"
    )
    assert (
        Perm((1, 2, 4, 0, 3, 5)).ascii_plot() == " | | | | | |\n"
        "-+-+-+-+-+-●-\n"
        " | | | | | |\n"
        "-+-+-●-+-+-+-\n"
        " | | | | | |\n"
        "-+-+-+-+-●-+-\n"
        " | | | | | |\n"
        "-+-●-+-+-+-+-\n"
        " | | | | | |\n"
        "-●-+-+-+-+-+-\n"
        " | | | | | |\n"
        "-+-+-+-●-+-+-\n"
        " | | | | | |"
    )
    for _ in range(10):
        perm = Perm.random(random.randint(0, 20))
        plot = perm.ascii_plot(cell_size=0).split("\n")
        for i in range(len(perm)):
            assert plot[len(perm) - perm[i] - 1][2 * i] == "\u25cf"


def test_cycle_notation():
    assert Perm().cycle_notation() == "( )"
    assert Perm((0,)).cycle_notation() == "( 0 )"
    assert Perm((0, 1)).cycle_notation() == "( 0 ) ( 1 )"
    assert Perm((7, 0, 1, 2, 5, 4, 3, 6)).cycle_notation() == "( 5 4 ) ( 7 6 3 2 1 0 )"



class Node(object):
    def __init__(self, value, prev=None, next=None):
        self.value = value
        self.prev = prev
        self.next = next

        if self.prev is not None:
            self.prev.next = self

        if self.next is not None:
            self.next.prev = self

    def __repr__(self):
        return "Node(%s)" % repr(self.value)


class DancingLinks(object):
    def __init__(self, lst=None):
        self.front = None
        self.back = None
        self.count = 0

        if lst is not None:
            for value in lst:
                self.append(value)

    def append(self, value):
        self.count += 1
        self.back = Node(value, self.back)
        if self.front is None:
            self.front = self.back

    def erase(self, node):
        self.count -= 1

        if node.prev is not None:
            node.prev.next = node.next

        if node.next is not None:
            node.next.prev = node.prev

        if node == self.front:
            self.front = node.next

        if node == self.back:
            self.back = node.prev

    def restore(self, node):
        self.count += 1

        if node.prev is not None:
            node.prev.next = node

        if node.next is not None:
            node.next.prev = node

        if node.next == self.front:
            self.front = node

        if node.prev == self.back:
            self.back = node

    def __len__(self):
        return self.count

    def __repr__(self):
        cur = self.front
        lst = []
        while cur is not None:
            lst.append(cur)
            cur = cur.next
        return "DancingLinks([%s])" % ", ".join(map(repr, lst))



import abc


class PermSetBase(metaclass=abc.ABCMeta):
    """Base class for all perm sets."""

    def contains(self, perm):
        # This comes for free when __contains__ is implemented
        return perm in self

    @abc.abstractmethod
    def of_length(self, perm):
        pass

    @abc.abstractmethod
    def __contains__(self, perm):
        pass

    @abc.abstractmethod
    def __getitem__(self, key):
        pass

    @abc.abstractmethod
    def __len__(self, key):
        pass

    def __str__(self):
        # Base __str__ which subclasses should override
        return "a perm set"

    def __repr__(self):
        # Base __repr__ which subclasses should override
        return "<A set of some perms>"



"""Return symmetries of sets."""


def rotate_90_clockwise_set(perms):
    try:
        return perms.__class__([p._rotate_right() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def rotate_180_clockwise_set(perms):
    try:
        return perms.__class__([p._rotate_180() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def rotate_270_clockwise_set(perms):
    try:
        return perms.__class__([p._rotate_left() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def rotate_set(perms):
    try:
        return perms.__class__([p.rotate() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def inverse_set(perms):
    try:
        return perms.__class__([p.inverse() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def reverse_set(perms):
    try:
        return perms.__class__([p.reverse() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def complement_set(perms):
    try:
        return perms.__class__([p.complement() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def antidiagonal_set(perms):
    try:
        return perms.__class__([p.flip_antidiagonal() for p in perms])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def all_symmetry_sets(input_perms):
    try:
        perms = sorted(tuple(input_perms))
        answer = set()
        for i in range(4):
            answer.add(tuple(sorted(perms)))
            answer.add(tuple(sorted(inverse_set(perms))))
            if i == 3:
                break
            perms = rotate_set(perms)
        return answer
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )


def lex_min(perms):
    try:
        return perms.__class__(sorted(all_symmetry_sets(perms))[0])
    except (AttributeError, TypeError):
        raise TypeError(
            ("perms parameter must be of type list, set, tuple of" " permuta.Perms")
        )



from permuta.meshpatt import MeshPatt
from permuta.perm import Perm


def is_sorted(w):
    """
    Return true if w is increasing, i.e., sorted

    EXAMPLES
    sage: is_sorted([1,3,2])
    False
    """

    n = len(w)

    if n <= 1:
        return True

    for i in range(n - 1):
        if w[i] > w[i + 1]:
            return False

    return True


#
# helper function for stack-sort and bubble-sort.
# finds the location of the maximal element
#


def loc_max(w):

    L = max(w)

    for i in range(L + 1):
        if w[i] == L:
            return i


#
# function takes a permutation w and does one pass of stack-sort on it
#


def stack_sort(w):

    i = len(w)

    if i == 0:
        return []

    if i == 1:
        return w
    else:
        [j, J] = [loc_max(w), max(w)]

        if j == 0:
            W2 = list(stack_sort(w[1:i]))
            W2.append(J)

            return W2

        if j == i - 1:
            W1 = list(stack_sort(w[0 : i - 1]))
            W1.append(J)

            return W1

        else:
            W1 = list(stack_sort(w[0:j]))
            W2 = list(stack_sort(w[j + 1 : i]))

            W1.extend(W2)
            W1.extend([J])

            return W1


#
# function takes a permutation w and does one pass of bubble-sort on it
#


def bubble_sort(w):

    i = len(w)

    if i == 0:
        return []

    if i == 1:
        return w
    else:
        [j, J] = [loc_max(w), max(w)]

        if j == 0:
            W2 = w[1:i]
            W2.append(J)

            return W2

        if j == i - 1:
            W1 = list(bubble_sort(w[0 : i - 1]))
            W1.append(J)

            return W1

        else:
            W1 = list(bubble_sort(w[0:j]))
            W2 = w[j + 1 : i]

            W1.extend(W2)
            W1.extend([J])

            return W1


def quick_sort_sub(perm):

    L = len(perm)

    if L == 0:
        return []

    sfps = list(Perm(perm).strong_fixed_points())
    if sfps:
        maxind = sfps[-1]
        return (
            quick_sort_sub([perm[i] for i in range(maxind)])
            + [perm[maxind]]
            + quick_sort_sub([perm[i] for i in range(maxind + 1, L)])
        )
    else:
        firstval = perm[0]
        smaller = filter(lambda x: x < firstval, perm)
        bigger = filter(lambda x: x > firstval, perm)
        return list(smaller) + [firstval] + list(bigger)


def quick_sort(perm):

    L = len(perm)

    if L == 0:
        return perm

    return Perm(quick_sort_sub(perm))


def BKV_sort(perm, B=[]):
    inp = list(perm)
    stackR = []  # the right stack read from top to bottom
    stackL = []  # the left stack read from top to bottom
    out = []
    while len(out) < len(perm):
        if inp and Perm.to_standard(stackR + [inp[0]]).reverse().avoids(*B):
            stackR.append(inp.pop(0))
        elif stackR and (
            Perm.to_standard(stackL + [stackR[-1]]).reverse().avoids(Perm((1, 0)))
        ):
            stackL.append(stackR.pop())
        elif stackL:
            out.append(stackL.pop())
        else:
            print("wtf")
            print(out, stackL, stackR, inp)
            assert False
    return Perm(out)


def stack_sortable(perm):
    return is_sorted(stack_sort(perm))


def quick_sortable(perm):
    return is_sorted(quick_sort(perm))


def BKV_sortable(perm, B=[]):
    return is_sorted(BKV_sort(perm, B))


def West_2_stack_sortable(perm):
    return is_sorted(stack_sort(stack_sort(perm)))


def West_3_stack_sortable(perm):
    return is_sorted(stack_sort(stack_sort(stack_sort(perm))))


def smooth(perm):
    return perm.avoids(Perm((0, 2, 1, 3)), Perm((1, 0, 3, 2)))


def forest_like(perm):
    return perm.avoids(Perm((0, 2, 1, 3)), MeshPatt(Perm((1, 0, 3, 2)), [(2, 2)]))


def Baxter(perm):
    return perm.avoids(
        MeshPatt(Perm((1, 3, 0, 2)), [(2, 2)]), MeshPatt(Perm((2, 0, 3, 1)), [(2, 2)])
    )


def SimSun(perm):
    return perm.avoids(MeshPatt(Perm((2, 1, 0)), [(1, 0), (1, 1), (2, 2)]))


def dihedral(perm):
    """
    We use the convention that D1 and D2 are not subgrroups of S1 and S2,
    respectively."""

    from sympy.combinatorics import Permutation
    from sympy.combinatorics.named_groups import DihedralGroup

    return Permutation(list(perm)) in DihedralGroup(len(perm))


def in_alternating_group(perm):
    """
    We use the convention that D1 and D2 are not subgrroups of S1 and S2,
    respectively."""

    from sympy.combinatorics import Permutation
    from sympy.combinatorics.named_groups import AlternatingGroup

    if perm == Perm(()):
        return True

    return Permutation(list(perm)) in AlternatingGroup(len(perm))


def perm_to_yt(perm):

    res = []

    def insert_in_row(i, k):
        if len(res) <= i:
            return res.append([k])

        cur_row = res[i]
        cur = cur_row[0]
        for ind, cur in enumerate(cur_row):
            if cur > k:
                cur_row[ind] = k
                return insert_in_row(i + 1, cur)
        else:
            cur_row.append(k)

    for k in perm:
        insert_in_row(0, k)

    return res


def tab_shape(tab):
    return list(map(len, tab))


def tableau_contains_shape(tab, sh):
    """
    Return True if the tableaux tab contains the shape sh
    """
    tab_sh = tab_shape(tab)

    lsh = len(sh)

    if len(tab_sh) < lsh:
        return False

    for i in range(lsh):
        if sh[i] > tab_sh[i]:
            return False

    return True


def yt_perm_avoids_22(perm):
    return not tableau_contains_shape(perm_to_yt(perm), [2, 2])


def yt_perm_avoids_32(perm):
    return not tableau_contains_shape(perm_to_yt(perm), [3, 2])


def av_231_and_mesh(perm):
    return perm.avoids(
        Perm((1, 2, 0)), MeshPatt(Perm((0, 1, 5, 2, 3, 4)), [(1, 6), (4, 5), (4, 6)])
    )


def hard_mesh(perm):
    return perm.avoids(
        MeshPatt(Perm((0, 1, 2)), [(0, 0), (1, 1), (2, 2), (3, 3)]),
        MeshPatt(Perm((0, 1, 2)), [(0, 3), (1, 2), (2, 1), (3, 0)]),
    )



import functools
import multiprocessing
import random

from .....descriptors.basis import AbstractBasis, Basis
from .....perm import Perm
from ....finite.permset_finite_specificlength import PermSetFiniteSpecificLength
from ....finite.permset_static import PermSetStatic
from ....unbounded.all.permset_all import PermSetAll
from ..permset_described import PermSetDescribed


class Avoiding(PermSetDescribed):
    """The base class for all avoidance classes."""

    # NOTE: Monkey patching of default subclass happens at end of file
    DESCRIPTOR_CLASS = AbstractBasis

    @property
    def basis(self):
        return self._descriptor

    def __hash__(self):
        return id(self)  # Requires the singleton property

    def __len__(self):
        raise NotImplementedError  # This is a hard task!

    def __repr__(self):
        return "Av({})".format(tuple(self.basis))

    def __str__(self):
        return "Av({})".format(", ".join(map(str, self.basis)))


class AvoidingGeneric(Avoiding):
    # Empty basis is dispatched to correct/another class (AvoidingEmpty)
    __CLASS_CACHE = {}
    _CACHE_LOCK = multiprocessing.Lock()

    def __new__(cls, basis):
        if basis in AvoidingGeneric.__CLASS_CACHE:
            return AvoidingGeneric.__CLASS_CACHE[basis]
        else:
            instance = super(AvoidingGeneric, cls).__new__(cls)
            # Generic case includes empty permutation
            instance.cache = [{Perm(): [0]}]
            AvoidingGeneric.__CLASS_CACHE[basis] = instance
            return instance

    def _ensure_level(self, level_number):
        # Ensure level is available
        while len(self.cache) <= level_number:
            nplusone = len(self.cache)  # really: len(perm) + 1
            if isinstance(self.basis, Basis):
                # Smart way when basis consists only of Perms
                # We will build the length n + 1 perms, from the length n perms

                n = nplusone - 1
                new_level = dict()
                max_size = max(len(p) for p in self.basis)
                last_level = self.cache[-1]

                # If we are currently building a length for which there is a
                #   basis element of that length, we set this to True
                check_length = nplusone in [len(b) for b in self.basis]
                # and get the basis element of this nplusone
                smaller_elems = [b for b in self.basis if len(b) == nplusone]

                def valid_insertions(perm):
                    res = None
                    for i in range(max(0, n - max_size), n):
                        val = perm[i]
                        subperm = perm.remove(i)
                        spots = self.cache[n - 1][subperm]
                        acceptable = [k for k in spots if k <= val] + [
                            k + 1 for k in spots if k >= val
                        ]
                        if res is None:
                            res = frozenset(acceptable)
                        res = res.intersection(acceptable)
                        if not res:
                            break
                    return res if res is not None else frozenset(range(nplusone))

                for perm in last_level.keys():
                    for value in valid_insertions(perm):
                        new_perm = perm.insert(index=nplusone, new_element=value)
                        if not check_length or new_perm not in smaller_elems:
                            new_level[new_perm] = []
                            last_level[perm].append(value)
            else:
                # Necessary non-smart way for e.g. MeshBasis
                new_level = set()
                for new_perm in PermSetAll().of_length(nplusone):
                    if new_perm.avoids(*self.basis):
                        new_level.add(new_perm)
            self.cache.append(new_level)

    def _get_level(self, level_number):
        with AvoidingGeneric._CACHE_LOCK:
            self._ensure_level(level_number)
        res = self.cache[level_number]
        if isinstance(res, dict):
            return self.cache[level_number].keys()
        return res

    def of_length(self, length):
        # TODO: Cache of instances?
        getter = functools.partial(self._get_level, length)
        return AvoidingGenericSpecificLength(length, self.basis, getter)

    def __getitem__(self, key):
        level_number = 0
        while True:
            level = self._get_level(level_number)
            if len(level) <= key:
                key -= len(level)
            else:  # TODO: So dumb
                return list(level)[key]
            level_number += 1

    def __next__(self):
        if self._iter is None:
            cached_perms = self._get_level(self._iter_number)
            if len(cached_perms) == 0:
                raise StopIteration
            self._iter = iter(cached_perms)
        try:
            return next(self._iter)
        except StopIteration:
            self._iter = None
            self._iter_number += 1
            return self.__next__()

    def __iter__(self):
        self._iter = None
        self._iter_number = 0
        return self

    def __contains__(self, perm):
        # TODO: Think about heuristics for switching to avoiding the patterns
        #       in the basis instead
        if isinstance(perm, Perm):
            length = len(perm)
            return perm in self._get_level(length)
        else:
            raise TypeError

    def is_subclass(self, other):
        """ Check if the `self` is a subclass of `other`. """
        return all(p1 not in self for p1 in other.basis)


class AvoidingGenericSpecificLength(PermSetFiniteSpecificLength):
    """Class for iterating through all perms of a specific length avoiding a
    basis."""

    # __slots__ = ("_length", "_basis", "_get_perms", "_iter")

    def __init__(self, length, basis, get_perms):
        self._length = length
        self._basis = basis
        self._get_perms = get_perms
        self._iter = None

    def of_length(self, length):
        if length != self._length:
            return PermSetStatic()
        else:
            return self

    def random(self):
        return random.choice(self._get_perms())

    def __contains__(self, other):
        """Check if other is a permutation in the set."""
        return isinstance(other, Perm) and other in self._get_perms()

    def __getitem__(self, key):
        raise NotImplementedError

    def __iter__(self):
        self._iter = iter(self._get_perms())
        return self

    def __len__(self):
        return len(self._get_perms())

    def __next__(self):
        return next(self._iter)

    def __str__(self):
        return "<PermSet of all perms of length {} avoiding {}>" "".format(
            self._length, self._basis
        )

    def __repr__(self):
        return "Av({}).of_length({})" "".format(repr(tuple(self._basis)), self._length)


# Set default Avoiding class to be dispatched
Avoiding.DEFAULT_CLASS = AvoidingGeneric



import argparse
import signal
import sys

from permuta import Av, Perm


def signal_handler(sig, frame):
    print()
    print("Exiting.")
    sys.exit(0)


def enumerate_class(args):
    basis = []
    for perm_str in args.basis.split("_"):
        perm_tuple = tuple(map(int, perm_str))
        if 0 not in perm_tuple:
            perm_tuple = tuple(i - 1 for i in perm_tuple)
        perm = Perm(perm_tuple, check=True)
        basis.append(perm)
    perm_class = Av(basis)
    n = 0
    print(f"Enumerating {perm_class}. Press Ctrl+C to exit.")
    signal.signal(signal.SIGINT, signal_handler)
    while True:
        num = len(list(perm_class.of_length(n)))
        print(f"{num}, ", end="", flush=True)
        n += 1


parser = argparse.ArgumentParser(description="A set of tools to work with permutations")
subparsers = parser.add_subparsers(title="subcommands")


# The count command
count_parser = subparsers.add_parser(
    "count", description="A tool to quickly get the enumeration of permutation classes"
)
count_parser.set_defaults(func=enumerate_class)
count_parser.add_argument(
    "basis",
    help="The basis as a string where the permutations are separated by '_' "
    "(e.g. '231_4321')",
)


def main():
    args = parser.parse_args()
    if not hasattr(args, "func"):
        parser.error("Invalid command")
    args.func(args)
    sys.exit(0)


if __name__ == "__main__":
    main()



import json
import os
import types
from collections import defaultdict

from permuta.bisc.bisc_subfunctions import (
    forb,
    mine,
    patterns_suffice_for_bad,
    patterns_suffice_for_good,
    run_clean_up,
    show_me,
    show_me_basis,
    to_sg_format,
)
from permuta.perm import Perm
from permuta.permset import PermSet


def bisc(A, m, n=None, report=False):

    if isinstance(A, list):
        D = defaultdict(list)
        for perm in A:
            D[len(perm)].append(perm)
    elif isinstance(A, types.FunctionType):
        if n is None:
            print("I will use permutations up to length 7")
            n = 7
        D = defaultdict(list)
        for i in range(n + 1):
            for perm in PermSet(i):
                if A(perm):
                    D[i].append(perm)

    elif isinstance(A, dict):
        D = A

    else:
        print(
            """BiSC can accept inputs:
        1. A list of permutations
        2. A property of permutations
        3. A dictionary of lengths pointing to permutations"""
        )
        assert False

    if n is None:
        n = max(D.keys())

    ci, goodpatts = mine(D, m, n, report=report)
    SG = forb(ci, goodpatts, m, report=report)

    return SG


def auto_bisc(prop):

    L = 8  # Want to sanity check on at least S8, and one above n
    n = 4
    m = 2

    if isinstance(prop, list):
        # If a list is passed in then we put it in a dictionary at the start
        A = defaultdict(list)
        # We will put the complement up to S8 into another dictionary
        B = defaultdict(list)
        for perm in prop:
            A[len(perm)].append(perm)
        if L not in A.keys():
            print("You should have permutations up to length at least 8")
            return
        for i in range(L + 1):
            for perm in PermSet(i):
                if perm not in A[i]:
                    B[i].append(perm)

    elif isinstance(prop, types.FunctionType):
        # If a property is passed in then we use it to populate both
        # both dictionaries up to S8
        A = defaultdict(list)
        B = defaultdict(list)
        for i in range(L + 1):
            A[i] = []
            B[i] = []
            for perm in PermSet(i):
                if prop(perm):
                    A[i].append(perm)
                else:
                    B[i].append(perm)

    elif (
        isinstance(prop, tuple)
        and isinstance(prop[0], dict)
        and isinstance(prop[1], dict)
    ):
        # If you already have the dictionaries you can pass them in as a tuple
        A = prop[0]
        B = prop[1]
        if L not in A.keys():
            print("You should have permutations up to length at least 8")
            return

    elif isinstance(prop, str):
        # If you pass in a string we assume it is pointing to a file in permsets
        print("Attempting to read perms from permsets")
        good_entry = None
        bad_entry = None
        with os.scandir("permsets") as entries:
            for i, entry in enumerate(entries):
                en = entry.name
                spl = en.split("_")
                if spl[0] == prop:
                    if (
                        good_entry is None
                        and spl[1] == "good"
                        and int(spl[2].split("len")[1]) >= L
                    ):
                        good_entry = en
                    if (
                        bad_entry is None
                        and spl[1] == "bad"
                        and int(spl[2].split("len")[1]) >= L
                    ):
                        bad_entry = en
                if good_entry is not None and bad_entry is not None:
                    break
        if good_entry is not None and bad_entry is not None:
            A = read_bisc_file("permsets/" + good_entry)
            B = read_bisc_file("permsets/" + bad_entry)
        else:
            print("The required files do not exist")
            return

    else:
        print(
            """BiSC can accept inputs:
        1. A list of permutations
        2. A property of permutations
        3. A dictionary of lengths pointing to permutations
        4. A string """
        )
        assert False

    while True:
        print("Learning patterns of length {} using perms of length {}".format(m, n))
        SG = bisc(A, m, n)
        show_me(SG, more=True)

        if SG != {}:
            print("Sanity checking the learned patterns up to length {}".format(L))
            val, avoiding_perms = patterns_suffice_for_bad(
                SG, L, B, stop_on_failure=True
            )
        else:
            val = False

        if val:
            print(
                "These patterns seem to suffice, I will now try to find a small basis"
            )

            ib = len(SG[min(SG.keys())].keys())
            while True:
                print("Trying to find a basis with at most {} patterns".format(ib))
                print("using perms of length {}".format(n))
                bases, dict_numbs_to_patts = run_clean_up(SG, B, n, limit_monitors=ib)

                if bases:
                    basis = bases[0]
                    show_me_basis(basis, dict_numbs_to_patts)

                    sg = to_sg_format(basis, dict_numbs_to_patts)

                    val, avoiding_perms = patterns_suffice_for_bad(
                        sg, L, B, stop_on_failure=True
                    )

                    if not val:
                        print("A bad basis was chosen.")
                        print("Increasing perm length to {}".format(n + 1))
                        n += 1
                        continue

                    val, containing_perms = patterns_suffice_for_good(
                        sg, L, A, stop_on_failure=True
                    )
                    if not val:
                        print("This is a bad basis. Need to learn from longer perms")
                        n += 1
                        break  # breaking out of the inner while-loop

                else:
                    print("No bases found. Increasing number of patterns in basis")
                    ib += 1
                    val = False

                if val:
                    print(
                        "!!! Found a basis with {} patts of length at most {}".format(
                            ib, m
                        )
                    )
                    print("!!! for the input using perms of length {}".format(n))
                    print("Basis: ", sg)
                    return sg

        else:
            print("Need to learn longer patterns")
            n += 1
            m += 1

        oldL = L
        if L < n + 1:
            L = n + 1
            if isinstance(prop, list) and L > max(A.keys):
                print("You need to input a longer list of permutations")
                return

            elif isinstance(prop, types.FunctionType):
                # Adding perms to the dictionaries
                for i in range(oldL + 1, L + 1):
                    print("Adding perms of length {}".format(i))
                    for perm in PermSet(i):
                        if prop(perm):
                            A[i].append(perm)
                        else:
                            B[i].append(perm)

            elif isinstance(prop, tuple) and L > min(max(A.keys()), max(B.keys())):
                print("You need to add longer permutations to the dictionaries")
                return

            elif isinstance(prop, str) and L > min(max(A.keys()), max(B.keys())):
                # If you pass in a string we assume it is pointing to a file in permsets
                print("Attempting to read perms from permsets")
                good_entry = None
                bad_entry = None
                with os.scandir("permsets") as entries:
                    for i, entry in enumerate(entries):
                        en = entry.name
                        spl = en.split("_")
                        if spl[0] == prop:
                            if (
                                good_entry is None
                                and spl[1] == "good"
                                and int(spl[2].split("len")[1]) >= L
                            ):
                                good_entry = en
                            if (
                                bad_entry is None
                                and spl[1] == "bad"
                                and int(spl[2].split("len")[1]) >= L
                            ):
                                bad_entry = en
                        if good_entry is not None and bad_entry is not None:
                            break
                if good_entry is not None and bad_entry is not None:
                    A = read_bisc_file("permsets/" + good_entry)
                    B = read_bisc_file("permsets/" + bad_entry)
                else:
                    print("The required files do not exist")
                    return


def create_bisc_input(N, prop):
    """
    Create a dictionary, D, containing keys 1, 2, 3, ..., N. Each key points to
    a list of permutations satisfying the property prop. The dictionary E has
    the same keys and they point to the complement.
    """

    A, B = {}, {}

    for n in range(N + 1):

        An, Bn = [], []

        for perm in PermSet(n):
            if prop(perm):
                An.append(perm)
            else:
                Bn.append(perm)

        A[n], B[n] = An, Bn

    return A, B


def write_bisc_files(N, prop, info):
    """
    Create a dictionary, D, containing keys 1, 2, 3, ..., N. Each key points to
    a list of permutations satisfying the property prop. The dictionary E has
    the same keys and they point to the complement.
    """

    A, B = {}, {}

    for n in range(N + 1):

        An, Bn = [], []

        for perm in PermSet(n):
            if prop(perm):
                An.append(perm)
            else:
                Bn.append(perm)

        A[n], B[n] = An, Bn
    f = open("{}_good_len{}".format(info, N), "a+")
    f.write(json.dumps(A))
    f.close()

    f = open("{}_bad_len{}".format(info, N), "a+")
    f.write(json.dumps(B))
    f.close()


def from_json(s):
    d = json.loads(s)
    return {int(key): list(map(Perm, values)) for key, values in d.items()}


def read_bisc_file(p):

    A = dict()

    f = open(p, "r")
    for line in f:
        A = from_json(line)
    f.close()

    return A



from permuta.misc import left_floor_and_ceiling, right_floor_and_ceiling


def test_left_floor_and_ceiling():
    iterable = [4, 5, 1, 2, 3, 6]
    expected = [
        (None, None),  # 4
        (0, None),  # 5
        (None, 0),  # 1
        (2, 0),  # 2
        (3, 0),  # 3
        (1, None),  # 6
    ]
    index = 0
    for fac in left_floor_and_ceiling(iterable):
        assert fac == expected[index]
        index += 1

    iterable = [4, 1, 2, 5, 3]
    expected = [(-1, 5), (-1, 0), (1, 0), (0, 5), (2, 0)]  # 4  # 1  # 2  # 5  # 3
    index = 0
    for fac in left_floor_and_ceiling(iterable, default_floor=-1, default_ceiling=5):
        assert fac == expected[index]
        index += 1

    iterable = [1, 2, 3]
    expected = [(None, None), (0, None), (1, None)]  # 1  # 2  # 3
    index = 0
    for fac in left_floor_and_ceiling(iterable):
        assert fac == expected[index]
        index += 1

    iterable = [3, 2, 1]
    expected = [(None, None), (None, 0), (None, 1)]  # 3  # 2  # 1
    index = 0
    for fac in left_floor_and_ceiling(iterable):
        assert fac == expected[index]
        index += 1


def test_right_floor_and_ceiling():
    iterable = [4, 5, 1, 2, 3, 6]
    expected = [
        (None, "on"),  # 4
        (0, "on"),  # 5
        (None, 0),  # 1
        (2, 0),  # 2
        (3, 0),  # 3
        (1, "on"),  # 6
    ]
    iterable = reversed(iterable)
    expected = list(reversed(expected))
    index = 0
    for fac in right_floor_and_ceiling(iterable, default_ceiling="on"):
        assert fac == expected[index]
        index += 1



import itertools


def cyclic_range(start, end, restart):
    """Yields start,...,end-1,restart,...,start-1."""
    return itertools.chain(range(start, end), range(restart, start))


def modulo_range(start, modulo):
    """Yields start,...,modulo-1,0,...,start-1."""
    return cyclic_range(start, modulo, 0)



from .algorithm_x import AlgorithmX


def exact_cover(bss, validcnt, max_cnt, ignore_first, allow_overlap_in_first):

    # curcover = []
    # ball = (1 << validcnt) - 1
    # care = ball & ~((1 << ignore_first) - 1)
    # def bt(at, left, done):
    #     if (done & care) == (ball & care):
    #         yield list(curcover)
    #     elif not (left == 0 or at == len(bss)):
    #         if (bss[at] & done &
    #             (care if allow_overlap_in_first else ball)) == 0:
    #             curcover.append(at)
    #             for res in bt(at + 1, left - 1, done | bss[at]):
    #                 yield res

    #             curcover.pop()

    #         for res in bt(at + 1, left, done):
    #             yield res

    # sols1 = []
    sols2 = []

    # for res in bt(0, max_cnt, 0):
    #     sols1.append(res)

    def handle_solution(sol):
        sols2.append(sol)
        return False

    ec = AlgorithmX(len(bss), validcnt - ignore_first, handle_solution)

    for i in range(len(bss)):
        bs = bss[i] >> ignore_first
        for j in range(validcnt - ignore_first):
            if (bs & (1 << j)) != 0:
                ec.set_value(i, j, True)

    ec.setup()
    ec.search(at_most=max_cnt)

    # print(sols1)
    # print(sols2)
    # assert sols1 == sols2

    return sols2


def exact_cover_smallest(bss, validcnt, max_cnt, ignore_first, allow_overlap_in_first):
    sols = []

    def handle_solution(sol):
        sols.append(sol)
        return False

    ec = AlgorithmX(len(bss), validcnt - ignore_first, handle_solution)
    for i in range(len(bss)):
        bs = bss[i] >> ignore_first
        for j in range(validcnt - ignore_first):
            if (bs & (1 << j)) != 0:
                ec.set_value(i, j, True)

    ec.setup()
    d = 1
    while max_cnt is None or d <= max_cnt:
        ec.can_continue = False
        ec.search(at_most=d)
        if sols or not ec.can_continue:
            break
        d += 1

    return sols



import abc

ABC = abc.ABCMeta("ABC", (object,), {})


class Shiftable(ABC):
    @abc.abstractmethod
    def shift_right(self, times=1):
        """Return self shifted times steps to the right.

        If shift is negative, shifted to the left.
        """
        pass

    @abc.abstractmethod
    def shift_left(self, times=1):
        """Return self shifted times steps to the left.

        If times is negative, shifted to the right.
        """
        pass

    @abc.abstractmethod
    def shift_up(self, times=1):
        """Return self shifted times steps up.

        If times is negative, shifted down.
        """
        pass

    @abc.abstractmethod
    def shift_down(self, times=1):
        """Return self shifted times steps down.

        If times is negative, shifted up.
        """
        pass
